# Copyright (c) 2022 The Choclo Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
import numpy as np
from numba import jit

from ..constants import VACUUM_MAGNETIC_PERMEABILITY


@jit(nopython=True)
def magnetic_field(
    easting_p,
    northing_p,
    upward_p,
    easting_q,
    northing_q,
    upward_q,
    magnetic_moment_east,
    magnetic_moment_north,
    magnetic_moment_up,
):
    r"""
    Magnetic field due to a dipole

    Returns the three components of the magnetic field due to a single dipole
    a single computation point.

    .. note::

        Use this function when all the three component of the magnetic fields
        are needed. Running this function is faster than computing each
        component separately. Use one of :func:`magnetic_e`,
        :func:`magnetic_n`, :func:`magnetic_u` if you need only one of them.

    Parameters
    ----------
    easting_p, northing_p, upward_p : float
        Easting, northing and upward coordinates of the observation point in
        meters.
    easting_q, northing_q, upward_q : float
        Easting, northing and upward coordinates of the dipole in meters.
    magnetic_moment_east, magnetic_moment_north, magnetic_moment_up : float
        The east, north and upward component of the magnetic moment vector of
        the dipole. Must be in :math:`A m^2`.

    Returns
    -------
    b_e, b_n, b_u : float
        Easting, northing and upward components of the magnetic field generated
        by the dipole on the observation point in :math:`\text{T}`.

    Notes
    -----
    Returns the three components of the magnetic field
    :math:`\mathbf{B}` on the observation point
    :math:`\mathbf{p} = (x_p, y_p, z_p)` generated by a single dipole located
    in :math:`\mathbf{q} = (x_q, y_q, z_q)` and magnetic moment
    :math:`\mathbf{m}=(m_x, m_y, m_z)`.

    .. math::

        \mathbf{B}(\mathbf{p}) =
            \frac{\mu_0}{4\pi}
            \left[
            \frac{
                3 (\mathbf{m} \cdot \mathbf{r}) \mathbf{r}
            }{
                \lVert r \rVert^5
            }
            -
            \frac{
                \mathbf{m}
            }{
                \lVert r \rVert^3
            }
            \right]

    where :math:`\mathbf{r} = \mathbf{p} - \mathbf{q}`,
    :math:`\lVert \cdot \rVert` refer to the :math:`L_2` norm
    and :math:`\mu_0` is the vacuum magnetic permeability.
    """
    r_e = easting_p - easting_q
    r_n = northing_p - northing_q
    r_u = upward_p - upward_q
    distance = np.sqrt(r_e**2 + r_n**2 + r_u**2)
    dotproduct = (
        magnetic_moment_east * r_e
        + magnetic_moment_north * r_n
        + magnetic_moment_up * r_u
    )
    c_m = VACUUM_MAGNETIC_PERMEABILITY / 4 / np.pi
    b_e = c_m * (
        3 * dotproduct * r_e / distance**5 - magnetic_moment_east / distance**3
    )
    b_n = c_m * (
        3 * dotproduct * r_n / distance**5 - magnetic_moment_north / distance**3
    )
    b_u = c_m * (3 * dotproduct * r_u / distance**5 - magnetic_moment_up / distance**3)
    return b_e, b_n, b_u


@jit(nopython=True)
def magnetic_e(
    easting_p,
    northing_p,
    upward_p,
    easting_q,
    northing_q,
    upward_q,
    magnetic_moment_east,
    magnetic_moment_north,
    magnetic_moment_up,
):
    r"""
    Easting component of the magnetic field due to a dipole

    Returns the easting component of the magnetic field by a single dipole on
    a single computation point

    Parameters
    ----------
    easting_p, northing_p, upward_p : float
        Easting, northing and upward coordinates of the observation point in
        meters.
    easting_q, northing_q, upward_q : float
        Easting, northing and upward coordinates of the dipole in meters.
    magnetic_moment_east, magnetic_moment_north, magnetic_moment_up : float
        The east, north and upward component of the magnetic moment vector of
        the dipole. Must be in :math:`A m^2`.

    Returns
    -------
    b_e : float
        Easting component of the magnetic field generated by the dipole
        on the observation point in :math:`\text{T}`.

    Notes
    -----
    Returns the easting component :math:`B_x(\mathbf{p})` of the magnetic field
    :math:`\mathbf{B}` on the observation point
    :math:`\mathbf{p} = (x_p, y_p, z_p)` generated by a single dipole located
    in :math:`\mathbf{q} = (x_q, y_q, z_q)` and magnetic moment
    :math:`\mathbf{m}=(m_x, m_y, m_z)`.

    .. math::

        B_x(\mathbf{p}) =
            \frac{\mu_0}{4\pi}
            \left[
            \frac{
                3 (\mathbf{m} \cdot \mathbf{r}) x
            }{
                \lVert r \rVert^5
            }
            -
            \frac{
                m_x
            }{
                \lVert r \rVert^3
            }
            \right]

    where :math:`\mathbf{r} = \mathbf{p} - \mathbf{q}`,
    :math:`\lVert \cdot \rVert` refer to the :math:`L_2` norm
    and :math:`\mu_0` is the vacuum magnetic permeability.
    """
    r_e = easting_p - easting_q
    r_n = northing_p - northing_q
    r_u = upward_p - upward_q
    distance = np.sqrt(r_e**2 + r_n**2 + r_u**2)
    dotproduct = (
        magnetic_moment_east * r_e
        + magnetic_moment_north * r_n
        + magnetic_moment_up * r_u
    )
    result = 3 * dotproduct * r_e / distance**5 - magnetic_moment_east / distance**3
    return VACUUM_MAGNETIC_PERMEABILITY / 4 / np.pi * result


@jit(nopython=True)
def magnetic_n(
    easting_p,
    northing_p,
    upward_p,
    easting_q,
    northing_q,
    upward_q,
    magnetic_moment_east,
    magnetic_moment_north,
    magnetic_moment_up,
):
    r"""
    Northing component of the magnetic field due to a dipole

    Returns the northing component of the magnetic field by a single dipole on
    a single computation point

    Parameters
    ----------
    easting_p, northing_p, upward_p : float
        Easting, northing and upward coordinates of the observation point in
        meters.
    easting_q, northing_q, upward_q : float
        Easting, northing and upward coordinates of the dipole in meters.
    magnetic_moment_east, magnetic_moment_north, magnetic_moment_up : float
        The east, north and upward component of the magnetic moment vector of
        the dipole. Must be in :math:`A m^2`.

    Returns
    -------
    b_n : float
        Northing component of the magnetic field generated by the dipole on the
        observation point in :math:`\text{T}`.

    Notes
    -----
    Returns the northing component :math:`B_y(\mathbf{p})` of the magnetic
    field :math:`\mathbf{B}` on the observation point
    :math:`\mathbf{p} = (x_p, y_p, z_p)` generated by a single dipole located
    in :math:`\mathbf{q} = (x_q, y_q, z_q)` and magnetic moment
    :math:`\mathbf{m}=(m_x, m_y, m_z)`.

    .. math::

        B_y(\mathbf{p}) =
            \frac{\mu_0}{4\pi}
            \left[
            \frac{
                3 (\mathbf{m} \cdot \mathbf{r}) y
            }{
                \lVert r \rVert^5
            }
            -
            \frac{
                m_y
            }{
                \lVert r \rVert^3
            }
            \right]

    where :math:`\mathbf{r} = \mathbf{p} - \mathbf{q}`,
    :math:`\lVert \cdot \rVert` refer to the :math:`L_2` norm
    and :math:`\mu_0` is the vacuum magnetic permeability.
    """
    r_e = easting_p - easting_q
    r_n = northing_p - northing_q
    r_u = upward_p - upward_q
    distance = np.sqrt(r_e**2 + r_n**2 + r_u**2)
    dotproduct = (
        magnetic_moment_east * r_e
        + magnetic_moment_north * r_n
        + magnetic_moment_up * r_u
    )
    result = 3 * dotproduct * r_n / distance**5 - magnetic_moment_north / distance**3
    return VACUUM_MAGNETIC_PERMEABILITY / 4 / np.pi * result


@jit(nopython=True)
def magnetic_u(
    easting_p,
    northing_p,
    upward_p,
    easting_q,
    northing_q,
    upward_q,
    magnetic_moment_east,
    magnetic_moment_north,
    magnetic_moment_up,
):
    r"""
    Upward component of the magnetic field due to a dipole

    Returns the upward component of the magnetic field by a single dipole on
    a single computation point

    Parameters
    ----------
    easting_p, northing_p, upward_p : float
        Easting, northing and upward coordinates of the observation point in
        meters.
    easting_q, northing_q, upward_q : float
        Easting, northing and upward coordinates of the dipole in meters.
    magnetic_moment_east, magnetic_moment_north, magnetic_moment_up : float
        The east, north and upward component of the magnetic moment vector of
        the dipole. Must be in :math:`A m^2`.

    Returns
    -------
    b_u : float
        Upward component of the magnetic field generated by the dipole on the
        observation point in :math:`\text{T}`.

    Notes
    -----
    Returns the upward component :math:`B_z(\mathbf{p})` of the magnetic
    field :math:`\mathbf{B}` on the observation point
    :math:`\mathbf{p} = (x_p, y_p, z_p)` generated by a single dipole located
    in :math:`\mathbf{q} = (x_q, y_q, z_q)` and magnetic moment
    :math:`\mathbf{m}=(m_x, m_y, m_z)`.

    .. math::

        B_z(\mathbf{p}) =
            \frac{\mu_0}{4\pi}
            \left[
            \frac{
                3 (\mathbf{m} \cdot \mathbf{r}) z
            }{
                \lVert r \rVert^5
            }
            -
            \frac{
                m_z
            }{
                \lVert r \rVert^3
            }
            \right]

    where :math:`\mathbf{r} = \mathbf{p} - \mathbf{q}`,
    :math:`\lVert \cdot \rVert` refer to the :math:`L_2` norm
    and :math:`\mu_0` is the vacuum magnetic permeability.
    """
    r_e = easting_p - easting_q
    r_n = northing_p - northing_q
    r_u = upward_p - upward_q
    distance = np.sqrt(r_e**2 + r_n**2 + r_u**2)
    dotproduct = (
        magnetic_moment_east * r_e
        + magnetic_moment_north * r_n
        + magnetic_moment_up * r_u
    )
    result = 3 * dotproduct * r_u / distance**5 - magnetic_moment_up / distance**3
    return VACUUM_MAGNETIC_PERMEABILITY / 4 / np.pi * result



# Copyright (c) 2022 The Choclo Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Kernels and forward modelling functions for point sources
"""
from ._forward import (
    gravity_e,
    gravity_ee,
    gravity_en,
    gravity_eu,
    gravity_n,
    gravity_nn,
    gravity_nu,
    gravity_pot,
    gravity_u,
    gravity_uu,
)
from ._kernels import (
    kernel_e,
    kernel_ee,
    kernel_en,
    kernel_eu,
    kernel_n,
    kernel_nn,
    kernel_nu,
    kernel_pot,
    kernel_u,
    kernel_uu,
)



# Copyright (c) 2022 The Choclo Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Test gravity forward modelling functions for rectangular prisms
"""
import numpy as np
import numpy.testing as npt
import pytest

from ..prism import (
    gravity_e,
    gravity_ee,
    gravity_en,
    gravity_eu,
    gravity_n,
    gravity_nn,
    gravity_nu,
    gravity_pot,
    gravity_u,
    gravity_uu,
)


@pytest.fixture(name="sample_prism_center")
def fixture_sample_prism_center():
    """
    Return the geometric center of the sample prism
    """
    return 30.5, 21.3, -43.5


@pytest.fixture(name="sample_density")
def fixture_sample_density():
    """
    Return the density for the sample prism
    """
    return 400


@pytest.fixture(name="sample_prism_dimensions")
def fixture_sample_prism_dimensions():
    """
    Return the dimensions of the sample prism
    """
    return 10, 15, 20


@pytest.fixture(name="sample_coordinate")
def fixture_sample_coordinate():
    """
    Define a sample observation point
    """
    return (16.7, 13.2, 7.8)


@pytest.fixture(name="sample_prism")
def fixture_sample_prism(sample_prism_center, sample_prism_dimensions):
    """
    Return a sample rectangular prism
    """
    west = sample_prism_center[0] - sample_prism_dimensions[0] / 2
    east = sample_prism_center[0] + sample_prism_dimensions[0] / 2
    south = sample_prism_center[1] - sample_prism_dimensions[1] / 2
    north = sample_prism_center[1] + sample_prism_dimensions[1] / 2
    bottom = sample_prism_center[2] - sample_prism_dimensions[2] / 2
    top = sample_prism_center[2] + sample_prism_dimensions[2] / 2
    prism = [west, east, south, north, bottom, top]
    return np.array(prism)


class TestSymmetryPotential:
    """
    Test the symmetry of gravity_pot of a rectangular prism
    """

    scalers = {
        "inside": 0.8,
        "surface": 1.0,
        "outside": 1.2,
    }

    @pytest.fixture(params=scalers.values(), ids=scalers.keys())
    def coords_in_vertices(self, sample_prism_center, sample_prism_dimensions, request):
        """
        Return observation points located in the vertices of the prism
        """
        # Get the coordinates of the sample prism center
        easting, northing, upward = sample_prism_center
        # Get the dimensions of the sample prism
        d_easting, d_northing, d_upward = sample_prism_dimensions
        # Get scaler
        scaler = request.param
        # Build the vertices
        vertices = list(
            [
                easting + i * scaler * d_easting / 2,
                northing + j * scaler * d_northing / 2,
                upward + k * scaler * d_upward / 2,
            ]
            for i in (-1, 1)
            for j in (-1, 1)
            for k in (-1, 1)
        )
        return vertices

    @pytest.fixture(params=scalers.values(), ids=scalers.keys())
    def coords_in_centers_of_easting_edges(
        self, sample_prism_center, sample_prism_dimensions, request
    ):
        """
        Return observation points located in the center of the prism edges
        parallel to easting axis
        """
        # Get the coordinates of the sample prism center
        easting, northing, upward = sample_prism_center
        # Get the dimensions of the sample prism
        d_easting, _, d_upward = sample_prism_dimensions
        # Get scaler
        scaler = request.param
        # Get the points in the symmetry group
        edges_easting = list(
            [
                easting + i * scaler * d_easting / 2,
                northing,
                upward + k * scaler * d_upward / 2,
            ]
            for i in (-1, 1)
            for k in (-1, 1)
        )
        return edges_easting

    @pytest.fixture(params=scalers.values(), ids=scalers.keys())
    def coords_in_centers_of_northing_edges(
        self, sample_prism_center, sample_prism_dimensions, request
    ):
        """
        Return observation points located in the center of the prism edges
        parallel to northing axis
        """
        # Get the coordinates of the sample prism center
        easting, northing, upward = sample_prism_center
        # Get the dimensions of the sample prism
        _, d_northing, d_upward = sample_prism_dimensions
        # Get scaler
        scaler = request.param
        # Get the points in the symmetry group
        edges_northing = list(
            [
                easting,
                northing + j * scaler * d_northing / 2,
                upward + k * scaler * d_upward / 2,
            ]
            for j in (-1, 1)
            for k in (-1, 1)
        )
        return edges_northing

    @pytest.fixture(params=scalers.values(), ids=scalers.keys())
    def coords_in_centers_of_upward_edges(
        self, sample_prism_center, sample_prism_dimensions, request
    ):
        """
        Return observation points located in the center of the prism edges
        parallel to upward axis
        """
        # Get the coordinates of the sample prism center
        easting, northing, upward = sample_prism_center
        # Get the dimensions of the sample prism
        d_easting, d_northing, _ = sample_prism_dimensions
        # Get scaler
        scaler = request.param
        # Get the points in the symmetry group
        edges_upward = list(
            [
                easting + i * scaler * d_easting / 2,
                northing + j * scaler * d_northing / 2,
                upward,
            ]
            for i in (-1, 1)
            for j in (-1, 1)
        )
        return edges_upward

    @pytest.fixture(params=scalers.values(), ids=scalers.keys())
    def coords_in_centers_of_easting_faces(
        self, sample_prism_center, sample_prism_dimensions, request
    ):
        """
        Return observation points located in the center of the faces normal to
        the easting axis
        """
        # Get the coordinates of the sample prism center
        easting, northing, upward = sample_prism_center
        # Get the dimensions of the sample prism
        d_easting, _, _ = sample_prism_dimensions
        # Get scaler
        scaler = request.param
        # Get the points in the symmetry group
        faces_normal_to_easting = [
            [easting - scaler * d_easting / 2, northing, upward],
            [easting + scaler * d_easting / 2, northing, upward],
        ]
        return faces_normal_to_easting

    @pytest.fixture(params=scalers.values(), ids=scalers.keys())
    def coords_in_centers_of_northing_faces(
        self, sample_prism_center, sample_prism_dimensions, request
    ):
        """
        Return observation points located in the center of the faces normal to
        the northing axis
        """
        # Get the coordinates of the sample prism center
        easting, northing, upward = sample_prism_center
        # Get the dimensions of the sample prism
        _, d_northing, _ = sample_prism_dimensions
        # Get scaler
        scaler = request.param
        # Get the points in the symmetry group
        faces_normal_to_northing = [
            [easting, northing - scaler * d_northing / 2, upward],
            [easting, northing + scaler * d_northing / 2, upward],
        ]
        return faces_normal_to_northing

    @pytest.fixture(params=scalers.values(), ids=scalers.keys())
    def coords_in_centers_of_upward_faces(
        self, sample_prism_center, sample_prism_dimensions, request
    ):
        """
        Return observation points located in the center of the faces normal to
        the northing axis
        """
        # Get the coordinates of the sample prism center
        easting, northing, upward = sample_prism_center
        # Get the dimensions of the sample prism
        _, _, d_upward = sample_prism_dimensions
        # Get scaler
        scaler = request.param
        # Get the points in the symmetry group
        faces_normal_to_upward = [
            [easting, northing, upward - scaler * d_upward / 2],
            [easting, northing, upward + scaler * d_upward / 2],
        ]
        return faces_normal_to_upward

    def test_vertices(self, coords_in_vertices, sample_prism, sample_density):
        """
        Test if gravity_pot satisfies symmetry on vertices
        """
        # Compute gravity_pot on every observation point of the symmetry group
        potential = list(
            gravity_pot(e, n, u, *sample_prism, sample_density)
            for e, n, u in coords_in_vertices
        )
        npt.assert_allclose(potential[0], potential)

    def test_centers_of_easting_edges(
        self, coords_in_centers_of_easting_edges, sample_prism, sample_density
    ):
        """
        Test if gravity_pot satisfies symmetry on centers of the edges
        parallel to the easting direction
        """
        # Compute gravity_pot on every observation point of the symmetry group
        potential = list(
            gravity_pot(e, n, u, *sample_prism, sample_density)
            for e, n, u in coords_in_centers_of_easting_edges
        )
        npt.assert_allclose(potential[0], potential)

    def test_centers_of_northing_edges(
        self, coords_in_centers_of_northing_edges, sample_prism, sample_density
    ):
        """
        Test if gravity_pot satisfies symmetry on centers of the edges
        parallel to the northing direction
        """
        # Compute gravity_pot on every observation point of the symmetry group
        potential = list(
            gravity_pot(e, n, u, *sample_prism, sample_density)
            for e, n, u in coords_in_centers_of_northing_edges
        )
        npt.assert_allclose(potential[0], potential)

    def test_centers_of_upward_edges(
        self, coords_in_centers_of_upward_edges, sample_prism, sample_density
    ):
        """
        Test if gravity_pot satisfies symmetry on centers of the edges
        parallel to the upward direction
        """
        # Compute gravity_pot on every observation point of the symmetry group
        potential = list(
            gravity_pot(e, n, u, *sample_prism, sample_density)
            for e, n, u in coords_in_centers_of_upward_edges
        )
        npt.assert_allclose(potential[0], potential)

    def test_centers_of_easting_faces(
        self, coords_in_centers_of_easting_faces, sample_prism, sample_density
    ):
        """
        Test if gravity_pot satisfies symmetry on centers of the
        centers of the faces normal to the easting direction
        """
        # Compute gravity_pot on every observation point of the symmetry group
        potential = list(
            gravity_pot(e, n, u, *sample_prism, sample_density)
            for e, n, u in coords_in_centers_of_easting_faces
        )
        npt.assert_allclose(potential[0], potential)

    def test_centers_of_northing_faces(
        self, coords_in_centers_of_northing_faces, sample_prism, sample_density
    ):
        """
        Test if gravity_pot satisfies symmetry on centers of the
        centers of the faces normal to the northing direction
        """
        # Compute gravity_pot on every observation point of the symmetry group
        potential = list(
            gravity_pot(e, n, u, *sample_prism, sample_density)
            for e, n, u in coords_in_centers_of_northing_faces
        )
        npt.assert_allclose(potential[0], potential)

    def test_centers_of_upward_faces(
        self, coords_in_centers_of_upward_faces, sample_prism, sample_density
    ):
        """
        Test if gravity_pot satisfies symmetry on centers of the
        centers of the faces normal to the upward direction
        """
        # Compute gravity_pot on every observation point of the symmetry group
        potential = list(
            gravity_pot(e, n, u, *sample_prism, sample_density)
            for e, n, u in coords_in_centers_of_upward_faces
        )
        npt.assert_allclose(potential[0], potential)


class TestSymmetryGravityE:
    """
    Test the symmetry of gravity_e of a rectangular prism
    """

    @pytest.fixture()
    def coords_in_northing_upward_plane(
        self, sample_prism_center, sample_prism_dimensions
    ):
        """
        Return observation points located in the northing-upward plane that
        pass along the center of the prism.
        """
        # Get the coordinates of the sample prism center
        center_easting, center_northing, center_upward = sample_prism_center
        # Get the dimensions of the sample prism
        _, d_northing, d_upward = sample_prism_dimensions
        # Build the points
        n_per_side = 5
        max_northing = d_northing / 2 * n_per_side
        max_upward = d_upward / 2 * n_per_side
        northing = np.linspace(-max_northing, max_northing, 2 * n_per_side + 1)
        upward = np.linspace(-max_upward, max_upward, 2 * n_per_side + 1)
        # Shift coordinates
        easting = np.zeros_like(northing) + center_easting
        northing += center_northing
        upward += center_upward
        return easting, northing, upward

    @pytest.fixture
    def mirrored_points(self, sample_prism_center, sample_prism_dimensions):
        """
        Define two set of mirrored points across the northing-upward plane
        that passes through the prism center
        """
        # Get the coordinates of the sample prism center
        center_easting, center_northing, center_upward = sample_prism_center
        # Get the dimensions of the sample prism
        d_easting, d_northing, d_upward = sample_prism_dimensions
        # Build the points
        n_per_side = 4
        max_easting = d_easting / 2 * n_per_side
        max_northing = d_northing / 2 * n_per_side
        max_upward = d_upward / 2 * n_per_side
        easting_east = np.linspace(d_easting / 2, max_easting, n_per_side)
        northing = np.linspace(-max_northing, max_northing, 2 * n_per_side + 1)
        upward = np.linspace(-max_upward, max_upward, 2 * n_per_side + 1)
        # Meshgrid
        easting_east, northing, upward = (
            array.ravel() for array in np.meshgrid(easting_east, northing, upward)
        )
        # Define easting_west
        easting_west = -easting_east
        # Shift coordinates
        easting_west += center_easting
        easting_east += center_easting
        northing += center_northing
        upward += center_upward
        return (easting_west, northing, upward), (easting_east, northing, upward)

    def test_northing_upward_plane(
        self,
        coords_in_northing_upward_plane,
        sample_prism,
        sample_prism_center,
        sample_density,
    ):
        """
        Test if gravity_e is null in points of the northing-upward plane that
        passes through its center.
        """
        # Compute gravity_e on every observation point of northing-upward plane
        g_e = list(
            gravity_e(e, n, u, *sample_prism, sample_density)
            for e, n, u in zip(*coords_in_northing_upward_plane)
        )
        # Compute gravity_e on a point slightly shifted from the prism center
        # (it will be our control for non-zero field)
        easting, northing, upward = sample_prism_center
        non_zero_g_e = gravity_e(
            easting + 1e-10, northing, upward, *sample_prism, sample_density
        )
        atol = np.abs(non_zero_g_e).max()
        npt.assert_allclose(g_e, 0, atol=atol)

    def test_mirrored_points(self, mirrored_points, sample_prism, sample_density):
        """
        Test if gravity_e is opposite in mirrored points across the
        northing-upward plane
        """
        west, east = mirrored_points
        # Compute gravity_e on every observation point of the two planes
        g_e_west = np.array(
            list(
                gravity_e(e, n, u, *sample_prism, sample_density)
                for e, n, u in zip(*west)
            )
        )
        g_e_east = np.array(
            list(
                gravity_e(e, n, u, *sample_prism, sample_density)
                for e, n, u in zip(*east)
            )
        )
        npt.assert_allclose(g_e_west, -g_e_east)

    @pytest.mark.parametrize(
        "density", (-200, 200), ids=("negative_density", "positive_density")
    )
    def test_sign(self, sample_prism_center, sample_prism, density):
        """
        Test that gravity_e has the correct sign on west and east points
        """
        # Get easting dimension of the prism
        d_easting = sample_prism[1] - sample_prism[0]
        # Define an observation point on the east of the prism
        point = (
            sample_prism_center[0] + d_easting,
            sample_prism_center[1],
            sample_prism_center[2],
        )
        # Compute gravity_e
        g_e = gravity_e(*point, *sample_prism, density)
        # Check if g_e has the opposite sign as its density (the acceleration
        # will point westward for a positive density, therefore negative
        # easting component)
        assert np.sign(g_e) != np.sign(density)

    def test_on_vertices(self, sample_prism, sample_density):
        """
        Test g_e on the vertices
        """
        # Get boundaries of the prism
        west, east, south, north, bottom, top = sample_prism
        # Split the vertices of the prism between the ones in the west and
        # east faces
        east_vertices = [[east, n, u] for n in (south, north) for u in (bottom, top)]
        west_vertices = [[west, n, u] for n in (south, north) for u in (bottom, top)]
        g_e_east = np.array(
            [
                gravity_e(*point, *sample_prism, sample_density)
                for point in east_vertices
            ]
        )
        g_e_west = np.array(
            [
                gravity_e(*point, *sample_prism, sample_density)
                for point in west_vertices
            ]
        )
        npt.assert_allclose(g_e_east, -g_e_west)


class TestSymmetryGravityN:
    """
    Test the symmetry of gravity_n of a rectangular prism
    """

    @pytest.fixture()
    def coords_in_easting_upward_plane(
        self, sample_prism_center, sample_prism_dimensions
    ):
        """
        Return observation points located in the easting-upward plane that
        pass along the center of the prism.
        """
        # Get the coordinates of the sample prism center
        center_easting, center_northing, center_upward = sample_prism_center
        # Get the dimensions of the sample prism
        d_easting, _, d_upward = sample_prism_dimensions
        # Build the points
        n_per_side = 5
        max_easting = d_easting / 2 * n_per_side
        max_upward = d_upward / 2 * n_per_side
        easting = np.linspace(-max_easting, max_easting, 2 * n_per_side + 1)
        upward = np.linspace(-max_upward, max_upward, 2 * n_per_side + 1)
        # Shift coordinates
        easting += center_easting
        northing = np.zeros_like(easting) + center_northing
        upward += center_upward
        return easting, northing, upward

    @pytest.fixture
    def mirrored_points(self, sample_prism_center, sample_prism_dimensions):
        """
        Define two set of mirrored points across the easting-upward plane
        that passes through the prism center
        """
        # Get the coordinates of the sample prism center
        center_easting, center_northing, center_upward = sample_prism_center
        # Get the dimensions of the sample prism
        d_easting, d_northing, d_upward = sample_prism_dimensions
        # Build the points
        n_per_side = 4
        max_easting = d_easting / 2 * n_per_side
        max_northing = d_northing / 2 * n_per_side
        max_upward = d_upward / 2 * n_per_side
        easting = np.linspace(-max_easting, max_easting, 2 * n_per_side + 1)
        northing_north = np.linspace(d_northing / 2, max_northing, n_per_side)
        upward = np.linspace(-max_upward, max_upward, 2 * n_per_side + 1)
        # Meshgrid
        easting, northing_north, upward = (
            array.ravel() for array in np.meshgrid(easting, northing_north, upward)
        )
        # Define northing_south
        northing_south = -northing_north
        # Shift coordinates
        easting += center_easting
        northing_north += center_northing
        northing_south += center_northing
        upward += center_upward
        return (easting, northing_south, upward), (easting, northing_north, upward)

    def test_northing_upward_plane(
        self,
        coords_in_easting_upward_plane,
        sample_prism,
        sample_prism_center,
        sample_density,
    ):
        """
        Test if gravity_n is null in points of the easting-upward plane that
        passes through its center.
        """
        # Compute gravity_n on every observation point of easting-upward plane
        g_n = list(
            gravity_n(e, n, u, *sample_prism, sample_density)
            for e, n, u in zip(*coords_in_easting_upward_plane)
        )
        # Compute gravity_n on a point slightly shifted from the prism center
        # (it will be our control for non-zero field)
        easting, northing, upward = sample_prism_center
        non_zero_g_n = gravity_n(
            easting, northing + 1e-10, upward, *sample_prism, sample_density
        )
        atol = np.abs(non_zero_g_n).max()
        npt.assert_allclose(g_n, 0, atol=atol)

    def test_mirrored_points(self, mirrored_points, sample_prism, sample_density):
        """
        Test if gravity_n is opposite in mirrored points across the
        easting-upward plane
        """
        south, north = mirrored_points
        # Compute gravity_n on every observation point of the two planes
        g_n_south = np.array(
            list(
                gravity_n(e, n, u, *sample_prism, sample_density)
                for e, n, u in zip(*south)
            )
        )
        g_n_north = np.array(
            list(
                gravity_n(e, n, u, *sample_prism, sample_density)
                for e, n, u in zip(*north)
            )
        )
        npt.assert_allclose(g_n_south, -g_n_north)

    @pytest.mark.parametrize(
        "density", (-200, 200), ids=("negative_density", "positive_density")
    )
    def test_sign(self, sample_prism_center, sample_prism, density):
        """
        Test that gravity_n has the correct sign on south and north points
        """
        # Get easting dimension of the prism
        d_northing = sample_prism[3] - sample_prism[2]
        # Define an observation point on the north of the prism
        point = (
            sample_prism_center[0],
            sample_prism_center[1] + d_northing,
            sample_prism_center[2],
        )
        # Compute gravity_e
        g_n = gravity_n(*point, *sample_prism, density)
        # Check if g_n has the opposite sign as its density (the acceleration
        # will point southward for a positive density, therefore negative
        # northing component)
        assert np.sign(g_n) != np.sign(density)

    def test_on_vertices(self, sample_prism, sample_density):
        """
        Test g_n on the vertices
        """
        # Get boundaries of the prism
        west, east, south, north, bottom, top = sample_prism
        # Split the vertices of the prism between the ones in the south and
        # north faces
        north_vertices = [[e, south, u] for e in (west, east) for u in (bottom, top)]
        south_vertices = [[e, north, u] for e in (west, east) for u in (bottom, top)]
        g_n_north = np.array(
            [
                gravity_n(*point, *sample_prism, sample_density)
                for point in north_vertices
            ]
        )
        g_n_south = np.array(
            [
                gravity_n(*point, *sample_prism, sample_density)
                for point in south_vertices
            ]
        )
        npt.assert_allclose(g_n_north, -g_n_south)


class TestSymmetryGravityU:
    """
    Test the symmetry of gravity_u of a rectangular prism
    """

    @pytest.fixture()
    def coords_in_easting_northing_plane(
        self, sample_prism_center, sample_prism_dimensions
    ):
        """
        Return observation points located in the easting-northing plane that
        pass along the center of the prism.
        """
        # Get the coordinates of the sample prism center
        center_easting, center_northing, center_upward = sample_prism_center
        # Get the dimensions of the sample prism
        d_easting, d_northing, _ = sample_prism_dimensions
        # Build the points
        n_per_side = 5
        max_easting = d_easting / 2 * n_per_side
        max_northing = d_northing / 2 * n_per_side
        easting = np.linspace(-max_easting, max_easting, 2 * n_per_side + 1)
        northing = np.linspace(-max_northing, max_northing, 2 * n_per_side + 1)
        # Shift coordinates
        easting += center_easting
        northing += center_northing
        upward = np.zeros_like(easting) + center_upward
        return easting, northing, upward

    @pytest.fixture
    def mirrored_points(self, sample_prism_center, sample_prism_dimensions):
        """
        Define two set of mirrored points across the easting-northing plane
        that passes through the prism center
        """
        # Get the coordinates of the sample prism center
        center_easting, center_northing, center_upward = sample_prism_center
        # Get the dimensions of the sample prism
        d_easting, d_northing, d_upward = sample_prism_dimensions
        # Build the points
        n_per_side = 4
        max_easting = d_easting / 2 * n_per_side
        max_northing = d_northing / 2 * n_per_side
        max_upward = d_upward / 2 * n_per_side
        easting = np.linspace(-max_easting, max_easting, 2 * n_per_side + 1)
        northing = np.linspace(-max_northing, max_northing, 2 * n_per_side + 1)
        upward_top = np.linspace(d_upward / 2, max_upward, n_per_side)
        # Meshgrid
        easting, northing, upward_top = (
            array.ravel() for array in np.meshgrid(easting, northing, upward_top)
        )
        # Define upward bottom
        upward_bottom = -upward_top
        # Shift coordinates
        easting += center_easting
        northing += center_northing
        upward_top += center_upward
        upward_bottom += center_upward
        return (easting, northing, upward_top), (easting, northing, upward_bottom)

    def test_easting_northing_plane(
        self,
        coords_in_easting_northing_plane,
        sample_prism,
        sample_prism_center,
        sample_density,
    ):
        """
        Test if gravity_u is null in points of the easting-northing plane that
        passes through its center.
        """
        # Compute gravity_u on every observation point of the easting-northing
        # plane
        g_u = list(
            gravity_u(e, n, u, *sample_prism, sample_density)
            for e, n, u in zip(*coords_in_easting_northing_plane)
        )
        # Compute gravity_u on a point slightly shifted from the prism center
        # (it will be our control for non-zero field)
        easting, northing, upward = sample_prism_center
        non_zero_g_u = gravity_u(
            easting, northing, upward + 1e-10, *sample_prism, sample_density
        )
        atol = np.abs(non_zero_g_u).max()
        npt.assert_allclose(g_u, 0, atol=atol)

    def test_mirrored_points(self, mirrored_points, sample_prism, sample_density):
        """
        Test if gravity_u is opposite in mirrored points across the
        easting-northing plane
        """
        top, bottom = mirrored_points
        # Compute gravity_u on every observation point of the two planes
        g_u_bottom = np.array(
            list(
                gravity_u(e, n, u, *sample_prism, sample_density)
                for e, n, u in zip(*bottom)
            )
        )
        g_u_top = np.array(
            list(
                gravity_u(e, n, u, *sample_prism, sample_density)
                for e, n, u in zip(*top)
            )
        )
        npt.assert_allclose(g_u_bottom, -g_u_top)

    @pytest.mark.parametrize(
        "density", (-200, 200), ids=("negative_density", "positive_density")
    )
    def test_sign(self, sample_prism_center, sample_prism, density):
        """
        Test that gravity_u has the correct sign on top and bottom points
        """
        # Get upward dimension of the prism
        d_upward = sample_prism[5] - sample_prism[4]
        # Define an observation point on top of the prism
        point = (
            sample_prism_center[0],
            sample_prism_center[1],
            sample_prism_center[2] + d_upward,
        )
        # Compute gravity_u
        g_u = gravity_u(*point, *sample_prism, density)
        # Check if g_u has the opposite sign as its density (the acceleration
        # will point downwards for a positive density, therefore negative
        # upward component)
        assert np.sign(g_u) != np.sign(density)

    def test_on_vertices(self, sample_prism, sample_density):
        """
        Test g_u on the vertices
        """
        # Get boundaries of the prism
        west, east, south, north, bottom, top = sample_prism
        # Split the vertices of the prism between the ones in the top and
        # bottom faces
        top_vertices = [[e, n, top] for e in (west, east) for n in (south, north)]
        bottom_vertices = [[e, n, bottom] for e in (west, east) for n in (south, north)]
        g_u_top = np.array(
            [gravity_u(*point, *sample_prism, sample_density) for point in top_vertices]
        )
        g_u_bottom = np.array(
            [
                gravity_u(*point, *sample_prism, sample_density)
                for point in bottom_vertices
            ]
        )
        npt.assert_allclose(g_u_top, -g_u_bottom)


class TestAccelerationFiniteDifferences:
    """
    Test acceleration components against finite-differences approximations of
    the gravity potential
    """

    # Define a small displacement for computing the finite differences
    delta = 0.0001  # 0.1 mm

    # Define expected relative error tolerance in the comparisons
    rtol = 1e-5

    @pytest.fixture
    def finite_diff_gravity_e(self, sample_coordinate, sample_prism, sample_density):
        """
        Compute gravity_e through finite differences of the gravity_pot
        """
        easting_p, northing_p, upward_p = sample_coordinate
        west, east = sample_prism[:2]
        # Compute shifted coordinate
        shifted_coordinate = (easting_p + self.delta, northing_p, upward_p)
        # Calculate g_e through finite differences
        g_e = (
            gravity_pot(*shifted_coordinate, *sample_prism, sample_density)
            - gravity_pot(*sample_coordinate, *sample_prism, sample_density)
        ) / self.delta
        return g_e

    @pytest.fixture
    def finite_diff_gravity_n(self, sample_coordinate, sample_prism, sample_density):
        """
        Compute gravity_n through finite differences of the gravity_pot
        """
        easting_p, northing_p, upward_p = sample_coordinate
        south, north = sample_prism[2:4]
        # Compute shifted coordinate
        shifted_coordinate = (easting_p, northing_p + self.delta, upward_p)
        # Calculate g_n through finite differences
        g_n = (
            gravity_pot(*shifted_coordinate, *sample_prism, sample_density)
            - gravity_pot(*sample_coordinate, *sample_prism, sample_density)
        ) / self.delta
        return g_n

    @pytest.fixture
    def finite_diff_gravity_u(self, sample_coordinate, sample_prism, sample_density):
        """
        Compute gravity_u through finite differences of the gravity_pot
        """
        easting_p, northing_p, upward_p = sample_coordinate
        bottom, top = sample_prism[-2:]
        # Compute shifted coordinate
        shifted_coordinate = (easting_p, northing_p, upward_p + self.delta)
        # Calculate g_u through finite differences
        g_u = (
            gravity_pot(*shifted_coordinate, *sample_prism, sample_density)
            - gravity_pot(*sample_coordinate, *sample_prism, sample_density)
        ) / self.delta
        return g_u

    def test_gravity_e(
        self, sample_coordinate, sample_prism, sample_density, finite_diff_gravity_e
    ):
        """
        Test gravity_e against finite differences of the gravity_pot
        """
        g_e = gravity_e(*sample_coordinate, *sample_prism, sample_density)
        npt.assert_allclose(g_e, finite_diff_gravity_e, rtol=self.rtol)

    def test_gravity_n(
        self, sample_coordinate, sample_prism, sample_density, finite_diff_gravity_n
    ):
        """
        Test gravity_n against finite differences of the gravity_pot
        """
        g_n = gravity_n(*sample_coordinate, *sample_prism, sample_density)
        npt.assert_allclose(g_n, finite_diff_gravity_n, rtol=self.rtol)

    def test_gravity_u(
        self, sample_coordinate, sample_prism, sample_density, finite_diff_gravity_u
    ):
        """
        Test gravity_u against finite differences of the gravity_pot
        """
        g_u = gravity_u(*sample_coordinate, *sample_prism, sample_density)
        npt.assert_allclose(g_u, finite_diff_gravity_u, rtol=self.rtol)


class TestTensorFiniteDifferences:
    """
    Test tensor components against finite-differences approximations of
    the gravity gradient
    """

    # Define a small displacement for computing the finite differences
    delta = 0.0001  # 0.1 mm

    # Define expected relative error tolerance in the comparisons
    rtol = 1e-5

    @pytest.fixture
    def finite_diff_gravity_ee(self, sample_coordinate, sample_prism, sample_density):
        """
        Compute gravity_ee through finite differences of the gravity_e
        """
        easting_p, northing_p, upward_p = sample_coordinate
        west, east = sample_prism[:2]
        # Compute shifted coordinate
        shifted_coordinate = (easting_p + self.delta, northing_p, upward_p)
        # Calculate g_ee through finite differences
        g_ee = (
            gravity_e(*shifted_coordinate, *sample_prism, sample_density)
            - gravity_e(*sample_coordinate, *sample_prism, sample_density)
        ) / self.delta
        return g_ee

    @pytest.fixture
    def finite_diff_gravity_nn(self, sample_coordinate, sample_prism, sample_density):
        """
        Compute gravity_nn through finite differences of the gravity_n
        """
        easting_p, northing_p, upward_p = sample_coordinate
        south, north = sample_prism[2:4]
        # Compute shifted coordinate
        shifted_coordinate = (easting_p, northing_p + self.delta, upward_p)
        # Calculate g_nn through finite differences
        g_nn = (
            gravity_n(*shifted_coordinate, *sample_prism, sample_density)
            - gravity_n(*sample_coordinate, *sample_prism, sample_density)
        ) / self.delta
        return g_nn

    @pytest.fixture
    def finite_diff_gravity_uu(self, sample_coordinate, sample_prism, sample_density):
        """
        Compute gravity_uu through finite differences of the gravity_u
        """
        easting_p, northing_p, upward_p = sample_coordinate
        bottom, top = sample_prism[-2:]
        # Compute shifted coordinate
        shifted_coordinate = (easting_p, northing_p, upward_p + self.delta)
        # Calculate g_u through finite differences
        g_uu = (
            gravity_u(*shifted_coordinate, *sample_prism, sample_density)
            - gravity_u(*sample_coordinate, *sample_prism, sample_density)
        ) / self.delta
        return g_uu

    @pytest.fixture
    def finite_diff_gravity_en(self, sample_coordinate, sample_prism, sample_density):
        """
        Compute gravity_en through finite differences of the gravity_e
        """
        easting_p, northing_p, upward_p = sample_coordinate
        west, east = sample_prism[:2]
        # Compute shifted coordinate
        shifted_coordinate = (easting_p, northing_p + self.delta, upward_p)
        # Calculate g_en through finite differences
        g_en = (
            gravity_e(*shifted_coordinate, *sample_prism, sample_density)
            - gravity_e(*sample_coordinate, *sample_prism, sample_density)
        ) / self.delta
        return g_en

    @pytest.fixture
    def finite_diff_gravity_eu(self, sample_coordinate, sample_prism, sample_density):
        """
        Compute gravity_eu through finite differences of the gravity_e
        """
        easting_p, northing_p, upward_p = sample_coordinate
        west, east = sample_prism[:2]
        # Compute shifted coordinate
        shifted_coordinate = (easting_p, northing_p, upward_p + self.delta)
        # Calculate g_en through finite differences
        g_eu = (
            gravity_e(*shifted_coordinate, *sample_prism, sample_density)
            - gravity_e(*sample_coordinate, *sample_prism, sample_density)
        ) / self.delta
        return g_eu

    @pytest.fixture
    def finite_diff_gravity_nu(self, sample_coordinate, sample_prism, sample_density):
        """
        Compute gravity_nu through finite differences of the gravity_n
        """
        easting_p, northing_p, upward_p = sample_coordinate
        west, east = sample_prism[:2]
        # Compute shifted coordinate
        shifted_coordinate = (easting_p, northing_p, upward_p + self.delta)
        # Calculate g_en through finite differences
        g_nu = (
            gravity_n(*shifted_coordinate, *sample_prism, sample_density)
            - gravity_n(*sample_coordinate, *sample_prism, sample_density)
        ) / self.delta
        return g_nu

    def test_gravity_ee(
        self, sample_coordinate, sample_prism, sample_density, finite_diff_gravity_ee
    ):
        """
        Test gravity_ee against finite differences of the gravity_e
        """
        g_ee = gravity_ee(*sample_coordinate, *sample_prism, sample_density)
        npt.assert_allclose(g_ee, finite_diff_gravity_ee, rtol=self.rtol)

    def test_gravity_nn(
        self, sample_coordinate, sample_prism, sample_density, finite_diff_gravity_nn
    ):
        """
        Test gravity_nn against finite differences of the gravity_n
        """
        g_nn = gravity_nn(*sample_coordinate, *sample_prism, sample_density)
        npt.assert_allclose(g_nn, finite_diff_gravity_nn, rtol=self.rtol)

    def test_gravity_uu(
        self, sample_coordinate, sample_prism, sample_density, finite_diff_gravity_uu
    ):
        """
        Test gravity_uu against finite differences of the gravity_u
        """
        g_uu = gravity_uu(*sample_coordinate, *sample_prism, sample_density)
        npt.assert_allclose(g_uu, finite_diff_gravity_uu, rtol=self.rtol)

    def test_gravity_en(
        self, sample_coordinate, sample_prism, sample_density, finite_diff_gravity_en
    ):
        """
        Test gravity_en against finite differences of the gravity_e
        """
        g_en = gravity_en(*sample_coordinate, *sample_prism, sample_density)
        npt.assert_allclose(g_en, finite_diff_gravity_en, rtol=self.rtol)

    def test_gravity_eu(
        self, sample_coordinate, sample_prism, sample_density, finite_diff_gravity_eu
    ):
        """
        Test gravity_eu against finite differences of the gravity_e
        """
        g_eu = gravity_eu(*sample_coordinate, *sample_prism, sample_density)
        npt.assert_allclose(g_eu, finite_diff_gravity_eu, rtol=self.rtol)

    def test_gravity_nu(
        self, sample_coordinate, sample_prism, sample_density, finite_diff_gravity_nu
    ):
        """
        Test gravity_nu against finite differences of the gravity_n
        """
        g_nu = gravity_nu(*sample_coordinate, *sample_prism, sample_density)
        npt.assert_allclose(g_nu, finite_diff_gravity_nu, rtol=self.rtol)


class TestLaplacian:
    @pytest.fixture
    def sample_observation_points(self, sample_prism_center, sample_prism_dimensions):
        """
        Return a 3D grid of observation points around the sample prism.
        The grid doesn't have points on edges and vertices (where the tensor
        components are not defined) and doesn't include the prism center where
        the fields will be zero.
        """
        # Get the coordinates of the sample prism center
        center_easting, center_northing, center_upward = sample_prism_center
        # Get the dimensions of the sample prism
        d_easting, d_northing, d_upward = sample_prism_dimensions
        # Build the points (avoid points in edges or vertices where the tensor
        # components are not defined)
        n_per_side = 4
        max_easting = 2 / 3 * d_easting * n_per_side
        max_northing = 2 / 3 * d_northing * n_per_side
        max_upward = 2 / 3 * d_upward * n_per_side
        easting = np.linspace(-max_easting, max_easting, 2 * n_per_side + 1)
        northing = np.linspace(-max_northing, max_northing, 2 * n_per_side + 1)
        upward = np.linspace(-max_upward, max_upward, 2 * n_per_side + 1)
        # Get the meshgrid
        easting, northing, upward = tuple(
            a.ravel() for a in np.meshgrid(easting, northing, upward)
        )
        # Remove the center of the prism
        is_prism_center = (easting == 0) & (northing == 0) & (upward == 0)
        easting, northing, upward = tuple(
            a[np.logical_not(is_prism_center)] for a in (easting, northing, upward)
        )
        # Shift the coordinates
        easting += center_easting
        northing += center_northing
        upward += center_upward
        return easting, northing, upward

    @pytest.mark.parametrize("left_component", ("g_ee", "g_nn", "g_uu"))
    def test_laplacian(
        self, sample_observation_points, sample_prism, sample_density, left_component
    ):
        """
        Test if diagonal tensor functions satisfy Laplace's equation
        """
        g_ee = np.array(
            [
                gravity_ee(e, n, u, *sample_prism, sample_density)
                for e, n, u in zip(*sample_observation_points)
            ]
        )
        g_nn = np.array(
            [
                gravity_nn(e, n, u, *sample_prism, sample_density)
                for e, n, u in zip(*sample_observation_points)
            ]
        )
        g_uu = np.array(
            [
                gravity_uu(e, n, u, *sample_prism, sample_density)
                for e, n, u in zip(*sample_observation_points)
            ]
        )
        if left_component == "g_ee":
            npt.assert_allclose(-g_ee, g_nn + g_uu)
        if left_component == "g_nn":
            npt.assert_allclose(-g_nn, g_ee + g_uu)
        if left_component == "g_uu":
            npt.assert_allclose(-g_uu, g_ee + g_nn)


class TestNonDiagonalTensor:
    """
    Test the behaviour of non-diagonal tensor components on critical points

    The non-diagonal tensor components evaluate the log function on
    ``log(x + r)``, where ``x`` is the shifted coordinate of the vertex and
    ``r`` is the distance between the vertex and the computation point. When
    the two other shifted coordinates are zero and ``x`` is negative, then
    ``x == -r``, making impossible to evaluate the log function.
    The ``_safe_log`` function accounts for this and safely evaluate the log
    function in these points.

    If we consider for example the ``g_en`` component, then observation points
    that fall above one of the vertices of the prism lead to this situation. If
    the ``_safe_log`` function is not properly defined, then the value of
    ``g_en`` on those points will be significantly different from the
    surrounding points, even with a different sign.

    The following test functions compare the values of the non-diagonal
    components on and around these particular observation points.
    """

    @pytest.fixture(name="prism")
    def prism(self):
        prism = [-10, 10, -20, 20, -50, -15]
        return np.array(prism, dtype=np.float64)

    @pytest.fixture(name="density")
    def density(self):
        return 400.0

    @pytest.mark.parametrize("easting_boundary", (0, 1))
    @pytest.mark.parametrize("northing_boundary", (2, 3))
    def test_g_en_above_vertices(
        self, prism, density, easting_boundary, northing_boundary
    ):
        """
        Test g_en on an observation point above the node and around it
        """
        vertex_easting = prism[easting_boundary]
        vertex_northing = prism[northing_boundary]
        # Consider an observation point above one of the vertex of the prism
        # and define a few observation points around it
        easting = np.linspace(vertex_easting - 3, vertex_easting + 3, 61)
        northing = np.linspace(vertex_northing - 3, vertex_northing + 3, 61)
        assert vertex_easting in easting and vertex_northing in northing
        upward = prism[5] + 1  # locate the observation points above the prism
        g_en = np.array(
            [
                gravity_en(e, n, upward, *prism, density)
                for e in easting
                for n in northing
            ]
        )
        # Check if all values in g_en have the same sign
        signs = np.sign(g_en)
        npt.assert_allclose(signs[0], signs)

    @pytest.mark.parametrize("easting_boundary", (0, 1))
    @pytest.mark.parametrize("upward_boundary", (4, 5))
    def test_g_eu_north_vertices(
        self, prism, density, easting_boundary, upward_boundary
    ):
        """
        Test g_eu on an observation point north the node and around it
        """
        vertex_easting = prism[easting_boundary]
        vertex_upward = prism[upward_boundary]
        # Consider an observation point at the north of one of the vertex of
        # the prism and define a few observation points around it
        easting = np.linspace(vertex_easting - 3, vertex_easting + 3, 61)
        upward = np.linspace(vertex_upward - 3, vertex_upward + 3, 61)
        assert vertex_easting in easting and vertex_upward in upward
        northing = prism[3] + 1  # locate the observation points north the prism
        g_eu = np.array(
            [
                gravity_eu(e, northing, u, *prism, density)
                for e in easting
                for u in upward
            ]
        )
        # Check if all values in g_eu have the same sign
        signs = np.sign(g_eu)
        npt.assert_allclose(signs[0], signs)

    @pytest.mark.parametrize("northing_boundary", (2, 3))
    @pytest.mark.parametrize("upward_boundary", (4, 5))
    def test_g_nu_east_vertices(
        self, prism, density, northing_boundary, upward_boundary
    ):
        """
        Test g_nu on an observation point north the node and around it
        """
        vertex_northing = prism[northing_boundary]
        vertex_upward = prism[upward_boundary]
        # Consider an observation point at the east of one of the vertex of
        # the prism and define a few observation points around it
        northing = np.linspace(vertex_northing - 3, vertex_northing + 3, 61)
        upward = np.linspace(vertex_upward - 3, vertex_upward + 3, 61)
        assert vertex_northing in northing and vertex_upward in upward
        easting = prism[1] + 1  # locate the observation points north the prism
        g_nu = np.array(
            [
                gravity_nu(easting, n, u, *prism, density)
                for n in northing
                for u in upward
            ]
        )
        # Check if all values in g_nu have the same sign
        signs = np.sign(g_nu)
        npt.assert_allclose(signs[0], signs)


class TestNonDiagonalTensorSymmetry:
    """
    Test symmetry of non-diagonal tensor components

    Each test will compute one tensor component on two regular grids. Each grid
    falls in one of two parallel planes that are equidistant to the prism. The
    grids will contain observation points that share a pair of coordinates
    with the vertices of the prism to test symmetry on those regions.
    For example, for `g_en` two horizontal grids will be defined: one above and
    one below the prism. The first one will contain points that fall right
    above its vertices, while the second one will contain points that fall
    right below them (these points share the easting and northing coordinates
    with these vertices).

    For the non-diagonal tensor components these points are where the
    modifications of the log functions take protagonism.
    """

    atol = 1e-18

    @pytest.fixture(name="prism")
    def prism(self):
        prism = [-10, 10, -20, 20, -30, 30]
        return np.array(prism, dtype=np.float64)

    @pytest.fixture(name="density")
    def density(self):
        return 400.0

    def test_g_en_symmetry(self, prism, density):
        """
        Test symmetry of g_en
        """
        west, east, south, north, bottom, top = prism[:]
        # Define two horizontal grids
        # (make sure that contain points that fall above and below the
        # vertices)
        easting = np.linspace(-40, 40, 41)
        northing = np.linspace(-40, 40, 41)
        assert west in easting and east in easting
        assert south in northing and north in northing
        delta = 2
        g_en_top = np.array(
            [
                gravity_en(e, n, top + delta, *prism, density)
                for e in easting
                for n in northing
            ]
        )
        g_en_bottom = np.array(
            [
                gravity_en(e, n, bottom - delta, *prism, density)
                for e in easting
                for n in northing
            ]
        )
        npt.assert_allclose(g_en_top, g_en_bottom, atol=self.atol)

    def test_g_eu_symmetry(self, prism, density):
        """
        Test symmetry of g_eu
        """
        west, east, south, north, bottom, top = prism[:]
        # Define two vertical grids parallel to the easting-upward plane
        # (make sure that contain points that fall north and south the
        # vertices)
        easting = np.linspace(-40, 40, 41)
        upward = np.linspace(-40, 40, 41)
        assert west in easting and east in easting
        assert bottom in upward and top in upward
        delta = 2
        g_eu_north = np.array(
            [
                gravity_eu(e, north + delta, u, *prism, density)
                for e in easting
                for u in upward
            ]
        )
        g_eu_south = np.array(
            [
                gravity_eu(e, south - delta, u, *prism, density)
                for e in easting
                for u in upward
            ]
        )
        npt.assert_allclose(g_eu_north, g_eu_south, atol=self.atol)

    def test_g_nu_symmetry(self, prism, density):
        """
        Test symmetry of g_nu
        """
        west, east, south, north, bottom, top = prism[:]
        # Define two vertical grids parallel to the northing-upward plane
        # (make sure that contain points that fall north and south the
        # vertices)
        northing = np.linspace(-40, 40, 41)
        upward = np.linspace(-40, 40, 41)
        assert south in northing and north in northing
        assert bottom in upward and top in upward
        delta = 2
        g_nu_north = np.array(
            [
                gravity_nu(east + delta, n, u, *prism, density)
                for n in northing
                for u in upward
            ]
        )
        g_nu_south = np.array(
            [
                gravity_nu(west - delta, n, u, *prism, density)
                for n in northing
                for u in upward
            ]
        )
        npt.assert_allclose(g_nu_north, g_nu_south, atol=self.atol)


class TestDiagonalTensorSingularities:
    """
    Test if diagonal tensor components behave as expected on their singular
    points

    Diagonal tensor components have singular points on:
    * prism vertices,
    * prism edges perpendicular to the tensor direction, and
    * prism faces normal to the tensor direction.

    For the first two cases, the forward modelling function should return
    ``np.nan``. For the last case, it should return the limit of the field when
    we approach from outside of the prism.
    """

    @pytest.fixture()
    def prism_boundaries(self):
        """
        Return the boundaries of the sample prism
        """
        west, east, south, north, bottom, top = -5.4, 10.1, 43.2, 79.5, -53.7, -44.3
        return west, east, south, north, bottom, top

    @pytest.mark.parametrize("function", (gravity_ee, gravity_nn, gravity_uu))
    def test_on_vertices(self, prism_boundaries, function):
        """
        Test if diagonal tensor components on vertices are equal to NaN
        """
        west, east, south, north, bottom, top = prism_boundaries
        prism = np.array([west, east, south, north, bottom, top])
        density = 1.0
        coordinates = tuple(
            a.ravel() for a in np.meshgrid([west, east], [south, north], [bottom, top])
        )
        results = list(
            function(e, n, u, *prism, density) for (e, n, u) in zip(*coordinates)
        )
        assert np.isnan(results).all()

    def test_gee_on_edges(self, prism_boundaries):
        """
        Test if gravity_ee on edges perpendicular to easting is equal to NaN
        """
        # Define prism
        west, east, south, north, bottom, top = prism_boundaries
        prism = np.array([west, east, south, north, bottom, top])
        density = 1.0
        # Define observation points on edges parallel to northing
        easting, upward = tuple(
            c.ravel() for c in np.meshgrid([west, east], [bottom, top])
        )
        northing = np.full_like(
            easting, (south + north) / 2
        )  # put points in the center of the edge
        results = list(
            gravity_ee(e, n, u, *prism, density)
            for (e, n, u) in zip(easting, northing, upward)
        )
        assert np.isnan(results).all()
        # Define observation points on edges parallel to upward
        easting, northing = tuple(
            c.ravel() for c in np.meshgrid([west, east], [south, north])
        )
        upward = np.full_like(
            easting, (bottom + top) / 2
        )  # put points in the center of the edge
        results = list(
            gravity_ee(e, n, u, *prism, density)
            for (e, n, u) in zip(easting, northing, upward)
        )
        assert np.isnan(results).all()

    def test_gnn_on_edges(self, prism_boundaries):
        """
        Test if gravity_nn on edges perpendicular to northing is equal to NaN
        """
        # Define prism
        west, east, south, north, bottom, top = prism_boundaries
        prism = np.array([west, east, south, north, bottom, top])
        density = 1.0
        # Define observation points on edges parallel to easting
        northing, upward = tuple(
            c.ravel() for c in np.meshgrid([south, north], [bottom, top])
        )
        easting = np.full_like(
            northing, (west + east) / 2
        )  # put points in the center of the edge
        results = list(
            gravity_nn(e, n, u, *prism, density)
            for (e, n, u) in zip(easting, northing, upward)
        )
        assert np.isnan(results).all()
        # Define observation points on edges parallel to upward
        easting, northing = tuple(
            c.ravel() for c in np.meshgrid([west, east], [south, north])
        )
        upward = np.full_like(
            easting, (bottom + top) / 2
        )  # put points in the center of the edge
        results = list(
            gravity_nn(e, n, u, *prism, density)
            for (e, n, u) in zip(easting, northing, upward)
        )
        assert np.isnan(results).all()

    def test_guu_on_edges(self, prism_boundaries):
        """
        Test if gravity_uu on edges perpendicular to upward is equal to NaN
        """
        # Define prism
        west, east, south, north, bottom, top = prism_boundaries
        prism = np.array([west, east, south, north, bottom, top])
        density = 1.0
        # Define observation points on edges parallel to easting
        northing, upward = tuple(
            c.ravel() for c in np.meshgrid([south, north], [bottom, top])
        )
        easting = np.full_like(
            northing, (west + east) / 2
        )  # put points in the center of the edge
        results = list(
            gravity_uu(e, n, u, *prism, density)
            for (e, n, u) in zip(easting, northing, upward)
        )
        assert np.isnan(results).all()
        # Define observation points on edges parallel to northing
        easting, upward = tuple(
            c.ravel() for c in np.meshgrid([west, east], [bottom, top])
        )
        northing = np.full_like(
            easting, (south + north) / 2
        )  # put points in the center of the edge
        results = list(
            gravity_ee(e, n, u, *prism, density)
            for (e, n, u) in zip(easting, northing, upward)
        )
        assert np.isnan(results).all()

    def test_gee_faces_symmetry(self, prism_boundaries):
        """
        Test if g_ee returns the same values on the faces normal to easting
        """
        # Define prism
        west, east, south, north, bottom, top = prism_boundaries
        prism = np.array([west, east, south, north, bottom, top])
        density = 1.0
        # Define observation points in the two faces normal to easting
        # (without including the edges)
        northing = np.linspace(south, north, 21)[1:-1]
        upward = np.linspace(bottom, top, 21)[1:-1]
        northing, upward = tuple(c.ravel() for c in np.meshgrid(northing, upward))
        east_face = np.full_like(northing, east)
        west_face = np.full_like(northing, west)
        result_east_face = list(
            gravity_ee(e, n, u, *prism, density)
            for (e, n, u) in zip(east_face, northing, upward)
        )
        result_west_face = list(
            gravity_ee(e, n, u, *prism, density)
            for (e, n, u) in zip(west_face, northing, upward)
        )
        npt.assert_allclose(result_east_face, result_west_face)

    def test_gee_faces_limit(self, prism_boundaries):
        """
        Test if g_ee on faces normal to easting returns the limit when
        approaching from the outside of the prism
        """
        # Define prism
        west, east, south, north, bottom, top = prism_boundaries
        prism = np.array([west, east, south, north, bottom, top])
        density = 1.0
        # Compute g_ee on the east face and on a slightly displaced point
        northing = (south + north) / 2
        upward = (bottom + top) / 2
        g_ee_on_face = gravity_ee(east, northing, upward, *prism, density)
        g_ee_close_to_face = gravity_ee(east + 1e-3, northing, upward, *prism, density)
        # Compare the results
        assert np.sign(g_ee_on_face) == np.sign(g_ee_close_to_face)

    def test_gnn_faces_limit(self, prism_boundaries):
        """
        Test if g_nn on faces normal to northing returns the limit when
        approaching from the outside of the prism
        """
        # Define prism
        west, east, south, north, bottom, top = prism_boundaries
        prism = np.array([west, east, south, north, bottom, top])
        density = 1.0
        # Compute g_nn on the north face and on a slightly displaced point
        easting = (west + east) / 2
        upward = (bottom + top) / 2
        g_nn_on_face = gravity_ee(easting, north, upward, *prism, density)
        g_nn_close_to_face = gravity_ee(easting, north + 1e-3, upward, *prism, density)
        # Compare the results
        assert np.sign(g_nn_on_face) == np.sign(g_nn_close_to_face)

    def test_guu_faces_limit(self, prism_boundaries):
        """
        Test if g_uu on faces normal to northing returns the limit when
        approaching from the outside of the prism
        """
        # Define prism
        west, east, south, north, bottom, top = prism_boundaries
        prism = np.array([west, east, south, north, bottom, top])
        density = 1.0
        # Compute g_uu on the top face and on a slightly displaced point
        easting = (west + east) / 2
        northing = (south + north) / 2
        g_uu_on_face = gravity_ee(easting, northing, top, *prism, density)
        g_uu_close_to_face = gravity_ee(easting, northing, top + 1e-3, *prism, density)
        # Compare the results
        assert np.sign(g_uu_on_face) == np.sign(g_uu_close_to_face)

    def test_gnn_faces_symmetry(self, prism_boundaries):
        """
        Test if g_nn returns the same values on the faces normal to northing
        """
        # Define prism
        west, east, south, north, bottom, top = prism_boundaries
        prism = np.array([west, east, south, north, bottom, top])
        density = 1.0
        # Define observation points in the two faces normal to northing
        # (without including the edges)
        easting = np.linspace(west, east, 21)[1:-1]
        upward = np.linspace(bottom, top, 21)[1:-1]
        easting, upward = tuple(c.ravel() for c in np.meshgrid(easting, upward))
        south_face = np.full_like(easting, south)
        north_face = np.full_like(easting, north)
        result_north_face = list(
            gravity_nn(e, n, u, *prism, density)
            for (e, n, u) in zip(easting, north_face, upward)
        )
        result_south_face = list(
            gravity_nn(e, n, u, *prism, density)
            for (e, n, u) in zip(easting, south_face, upward)
        )
        npt.assert_allclose(result_north_face, result_south_face)

    def test_guu_faces_symmetry(self, prism_boundaries):
        """
        Test if g_uu returns the same values on the horizontal faces
        """
        # Define prism
        west, east, south, north, bottom, top = prism_boundaries
        prism = np.array([west, east, south, north, bottom, top])
        density = 1.0
        # Define observation points in the two horizontal faces
        # (without including the edges)
        easting = np.linspace(west, east, 21)[1:-1]
        northing = np.linspace(south, north, 21)[1:-1]
        easting, northing = tuple(c.ravel() for c in np.meshgrid(easting, northing))
        top_face = np.full_like(easting, top)
        bottom_face = np.full_like(easting, bottom)
        result_top_face = list(
            gravity_uu(e, n, u, *prism, density)
            for (e, n, u) in zip(easting, northing, top_face)
        )
        result_bottom_face = list(
            gravity_uu(e, n, u, *prism, density)
            for (e, n, u) in zip(easting, northing, bottom_face)
        )
        npt.assert_allclose(result_top_face, result_bottom_face)


class TestNonDiagonalTensorSingularities:
    """
    Test if non-diagonal tensor components behave as expected on their singular
    points

    Non-diagonal tensor components have singular points on:
    * prism vertices, and
    * prism edges perpendicular to the two tensor directions.

    In both cases, the forward modelling function should return ``np.nan``.
    """

    @pytest.fixture()
    def prism_boundaries(self):
        """
        Return the boundaries of the sample prism
        """
        west, east, south, north, bottom, top = -5.4, 10.1, 43.2, 79.5, -53.7, -44.3
        return west, east, south, north, bottom, top

    @pytest.mark.parametrize("function", (gravity_en, gravity_eu, gravity_nu))
    def test_on_vertices(self, prism_boundaries, function):
        """
        Test if non-diagonal tensor components on vertices are equal to NaN
        """
        west, east, south, north, bottom, top = prism_boundaries
        prism = np.array([west, east, south, north, bottom, top])
        density = 1.0
        coordinates = tuple(
            a.ravel() for a in np.meshgrid([west, east], [south, north], [bottom, top])
        )
        results = list(
            function(e, n, u, *prism, density) for (e, n, u) in zip(*coordinates)
        )
        assert np.isnan(results).all()

    def test_gen_edges(self, prism_boundaries):
        """
        Test if gravity_en returns np.nan on the vertical edges
        """
        # Define prism
        west, east, south, north, bottom, top = prism_boundaries
        prism = np.array([west, east, south, north, bottom, top])
        density = 1.0
        # Define observation points in the vertical edges of the prism
        easting, northing = tuple(
            c.ravel() for c in np.meshgrid([west, east], [south, north])
        )
        upward = np.full_like(
            easting, (bottom + top) / 2
        )  # put points in the center of the edge
        results = list(
            gravity_en(e, n, u, *prism, density)
            for (e, n, u) in zip(easting, northing, upward)
        )
        assert np.isnan(results).all()

    def test_geu_edges(self, prism_boundaries):
        """
        Test if gravity_eu returns np.nan on the edges parallel to northing
        """
        # Define prism
        west, east, south, north, bottom, top = prism_boundaries
        prism = np.array([west, east, south, north, bottom, top])
        density = 1.0
        # Define observation points on edges parallel to northing
        easting, upward = tuple(
            c.ravel() for c in np.meshgrid([west, east], [bottom, top])
        )
        northing = np.full_like(
            easting, (south + north) / 2
        )  # put points in the center of the edge
        results = list(
            gravity_eu(e, n, u, *prism, density)
            for (e, n, u) in zip(easting, northing, upward)
        )
        assert np.isnan(results).all()

    def test_gnu_edges(self, prism_boundaries):
        """
        Test if gravity_nu returns np.nan on the edges parallel to easting
        """
        # Define prism
        west, east, south, north, bottom, top = prism_boundaries
        prism = np.array([west, east, south, north, bottom, top])
        density = 1.0
        # Define observation points on edges parallel to easting
        northing, upward = tuple(
            c.ravel() for c in np.meshgrid([south, north], [bottom, top])
        )
        easting = np.full_like(
            northing, (west + east) / 2
        )  # put points in the center of the edge
        results = list(
            gravity_nu(e, n, u, *prism, density)
            for (e, n, u) in zip(easting, northing, upward)
        )
        assert np.isnan(results).all()



# Copyright (c) 2022 The Choclo Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Gravity forward modelling functions for rectangular prisms
"""
import numpy as np
from numba import jit

from ..constants import GRAVITATIONAL_CONST
from ._kernels import (
    kernel_e,
    kernel_ee,
    kernel_en,
    kernel_eu,
    kernel_n,
    kernel_nn,
    kernel_nu,
    kernel_pot,
    kernel_u,
    kernel_uu,
)
from ._utils import (
    is_point_on_east_face,
    is_point_on_easting_edge,
    is_point_on_north_face,
    is_point_on_northing_edge,
    is_point_on_top_face,
    is_point_on_upward_edge,
)


@jit(nopython=True)
def gravity_pot(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    density,
):
    r"""
    Gravitational potential field due to a rectangular prism

    Returns the gravitational potential field produced by a single rectangular
    prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    density : float
        Density of the rectangular prism in kilograms per cubic meter.

    Returns
    -------
    potential : float
        Gravitational potential field generated by the rectangular prism on the
        observation point in :math:`\text{J}/\text{kg}`.

    Notes
    -----
    Returns the gravitational potential field :math:`V(\mathbf{p})` on the
    observation point :math:`\mathbf{p} = (x_p, y_p, z_p)` generated by
    a single rectangular prism defined by its boundaries :math:`x_1, x_2, y_1,
    y_2, z_1, z_2` and with a density :math:`\rho`:

    .. math::

        V(\mathbf{p}) =
            G
            \rho \,\,
            \Bigg\lvert \Bigg\lvert \Bigg\lvert
            k_V(x, y, z)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}

    where

    .. math::

        k_V(x, y, z) &=
            x y \, \operatorname{safe-ln} (z, r)
            + y z \, \operatorname{safe-ln} (x, r)
            + z x \, \operatorname{safe-ln} (y, r) \\
            - \frac{x^2}{2} &\operatorname{safe-arctan} \left( yz, xr \right)
            - \frac{y^2}{2} \operatorname{safe-arctan} \left( zx, yr \right)
            - \frac{z^2}{2} \operatorname{safe-arctan} \left( xy, zr \right),

    .. math::

        r = \sqrt{x^2 + y^2 + z^2},

    and

    .. math::

        X_1 = x_1 - x_p \\
        X_2 = x_2 - x_p \\
        Y_1 = y_1 - y_p \\
        Y_2 = y_2 - y_p \\
        Z_1 = z_1 - z_p \\
        Z_2 = z_2 - z_p

    are the shifted coordinates of the prism boundaries and :math:`G` is the
    Universal Gravitational Constant.

    The :math:`\operatorname{safe-ln}` and :math:`\operatorname{safe-arctan}`
    functions are defined as follows:

    .. math::

        \operatorname{safe-ln}(x, r) =
        \begin{cases}
            0 & x = 0, r = 0 \\
            \ln(x + r) & x \ge 0 \\
            \ln((r^2 - x^2) / (r - x)) & x < 0, r \ne -x \\
            -\ln(-2 x) & x < 0, r = -x
        \end{cases}

    .. math::

        \operatorname{safe-arctan} \left( y, x \right) =
        \begin{cases}
            \text{arctan}\left( \frac{y}{x} \right) & x \ne 0 \\
            \frac{\pi}{2} & x = 0 \quad \text{and} \quad y > 0 \\
            -\frac{\pi}{2} & x = 0 \quad \text{and} \quad y < 0 \\
            0 & x = 0 \quad \text{and} \quad y = 0 \\
        \end{cases}

    These were defined after [Fukushima2020]_ and guarantee a good accuracy on
    any observation point.

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    return (
        GRAVITATIONAL_CONST
        * density
        * _evaluate_kernel(
            easting,
            northing,
            upward,
            prism_west,
            prism_east,
            prism_south,
            prism_north,
            prism_bottom,
            prism_top,
            kernel_pot,
        )
    )


@jit(nopython=True)
def gravity_e(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    density,
):
    r"""
    Easting component of the gravitational acceleration due to a prism

    Returns the easting component of the gravitational acceleration produced by
    a single rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    density : float
        Density of the rectangular prism in kilograms per cubic meter.

    Returns
    -------
    g_e : float
        Easting component of the gravitational acceleration generated by the
        rectangular prism on the observation point in
        :math:`\text{m}/\text{s}^2`.

    Notes
    -----
    Returns the easting component :math:`g_x(\mathbf{p})` of the gravitational
    acceleration :math:`\mathbf{g}` on the observation point :math:`\mathbf{p}
    = (x_p, y_p, z_p)` generated by a single rectangular prism defined by its
    boundaries :math:`x_1, x_2, y_1, y_2, z_1, z_2` and with a density
    :math:`\rho`:

    .. math::

        g_x(\mathbf{p}) =
            G
            \rho \,\,
            \Bigg\lvert \Bigg\lvert \Bigg\lvert
            k_x(x, y, z)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}

    where

    .. math::

        k_x(x, y, z) =
            \left[
            y \, \operatorname{safe-ln} (z, r)
            + z \, \operatorname{safe-ln} (y, r)
            - x \, \operatorname{safe-arctan} \left( yz, xr \right)
            \right]

    .. math::

        r = \sqrt{x^2 + y^2 + z^2},

    and

    .. math::

        X_1 = x_1 - x_p \\
        X_2 = x_2 - x_p \\
        Y_1 = y_1 - y_p \\
        Y_2 = y_2 - y_p \\
        Z_1 = z_1 - z_p \\
        Z_2 = z_2 - z_p

    are the shifted coordinates of the prism boundaries and :math:`G` is the
    Universal Gravitational Constant.

    The :math:`\operatorname{safe-ln}` and :math:`\operatorname{safe-arctan}`
    functions are defined as follows:

    .. math::

        \operatorname{safe-ln}(x, r) =
        \begin{cases}
            0 & x = 0, r = 0 \\
            \ln(x + r) & x \ge 0 \\
            \ln((r^2 - x^2) / (r - x)) & x < 0, r \ne -x \\
            -\ln(-2 x) & x < 0, r = -x
        \end{cases}

    .. math::

        \operatorname{safe-arctan} \left( y, x \right) =
        \begin{cases}
            \text{arctan}\left( \frac{y}{x} \right) & x \ne 0 \\
            \frac{\pi}{2} & x = 0 \quad \text{and} \quad y > 0 \\
            -\frac{\pi}{2} & x = 0 \quad \text{and} \quad y < 0 \\
            0 & x = 0 \quad \text{and} \quad y = 0 \\
        \end{cases}

    These were defined after [Fukushima2020]_ and guarantee a good accuracy on
    any observation point.

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    return (
        GRAVITATIONAL_CONST
        * density
        * _evaluate_kernel(
            easting,
            northing,
            upward,
            prism_west,
            prism_east,
            prism_south,
            prism_north,
            prism_bottom,
            prism_top,
            kernel_e,
        )
    )


@jit(nopython=True)
def gravity_n(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    density,
):
    r"""
    Northing component of the gravitational acceleration due to a prism

    Returns the northing component of the gravitational acceleration produced
    by a single rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    density : float
        Density of the rectangular prism in kilograms per cubic meter.

    Returns
    -------
    g_n : float
        Northing component of the gravitational acceleration generated by the
        rectangular prism on the observation point in
        :math:`\text{m}/\text{s}^2`.

    Notes
    -----
    Returns the northing component :math:`g_y(\mathbf{p})` of the gravitational
    acceleration :math:`\mathbf{g}` on the observation point :math:`\mathbf{p}
    = (x_p, y_p, z_p)` generated by a single rectangular prism defined by its
    boundaries :math:`x_1, x_2, y_1, y_2, z_1, z_2` and with a density
    :math:`\rho`:

    .. math::

        g_y(\mathbf{p}) =
            G
            \rho \,\,
            \Bigg\lvert \Bigg\lvert \Bigg\lvert
            k_y(x, y, z)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}

    where

    .. math::

        k_y(x, y, z) =
            \left[
            z \, \operatorname{safe-ln} (x, r)
            + x \, \operatorname{safe-ln} (z, r)
            - y \, \operatorname{safe-arctan} \left( zx, yr \right)
            \right]

    .. math::

        r = \sqrt{x^2 + y^2 + z^2},

    and

    .. math::

        X_1 = x_1 - x_p \\
        X_2 = x_2 - x_p \\
        Y_1 = y_1 - y_p \\
        Y_2 = y_2 - y_p \\
        Z_1 = z_1 - z_p \\
        Z_2 = z_2 - z_p

    are the shifted coordinates of the prism boundaries and :math:`G` is the
    Universal Gravitational Constant.

    The :math:`\operatorname{safe-ln}` and :math:`\operatorname{safe-arctan}`
    functions are defined as follows:

    .. math::

        \operatorname{safe-ln}(x, r) =
        \begin{cases}
            0 & x = 0, r = 0 \\
            \ln(x + r) & x \ge 0 \\
            \ln((r^2 - x^2) / (r - x)) & x < 0, r \ne -x \\
            -\ln(-2 x) & x < 0, r = -x
        \end{cases}

    .. math::

        \operatorname{safe-arctan} \left( y, x \right) =
        \begin{cases}
            \text{arctan}\left( \frac{y}{x} \right) & x \ne 0 \\
            \frac{\pi}{2} & x = 0 \quad \text{and} \quad y > 0 \\
            -\frac{\pi}{2} & x = 0 \quad \text{and} \quad y < 0 \\
            0 & x = 0 \quad \text{and} \quad y = 0 \\
        \end{cases}

    These were defined after [Fukushima2020]_ and guarantee a good accuracy on
    any observation point.

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    return (
        GRAVITATIONAL_CONST
        * density
        * _evaluate_kernel(
            easting,
            northing,
            upward,
            prism_west,
            prism_east,
            prism_south,
            prism_north,
            prism_bottom,
            prism_top,
            kernel_n,
        )
    )


@jit(nopython=True)
def gravity_u(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    density,
):
    r"""
    Upward component of the gravitational acceleration due to a prism

    Returns the upward component of the gravitational acceleration produced by
    a single rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    density : float
        Density of the rectangular prism in kilograms per cubic meter.

    Returns
    -------
    g_u : float
        Upward component of the gravitational acceleration generated by the
        rectangular prism on the observation point in
        :math:`\text{m}/\text{s}^2`.

    Notes
    -----
    Returns the upward component :math:`g_z(\mathbf{p})` of the gravitational
    acceleration :math:`\mathbf{g}` on the observation point :math:`\mathbf{p}
    = (x_p, y_p, z_p)` generated by a single rectangular prism defined by its
    boundaries :math:`x_1, x_2, y_1, y_2, z_1, z_2` and with a density
    :math:`\rho`:

    .. math::

        g_z(\mathbf{p}) =
            G
            \rho \,\,
            \Bigg\lvert \Bigg\lvert \Bigg\lvert
            k_z(x, y, z)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}

    where

    .. math::

        k_z(x, y, z) =
            - \left[
            x \, \operatorname{safe-ln} (y, r)
            + y \, \operatorname{safe-ln} (x, r)
            - z \, \operatorname{safe-arctan} \left( xy, zr \right)
            \right]

    .. math::

        r = \sqrt{x^2 + y^2 + z^2},

    and

    .. math::

        X_1 = x_1 - x_p \\
        X_2 = x_2 - x_p \\
        Y_1 = y_1 - y_p \\
        Y_2 = y_2 - y_p \\
        Z_1 = z_1 - z_p \\
        Z_2 = z_2 - z_p

    are the shifted coordinates of the prism boundaries and :math:`G` is the
    Universal Gravitational Constant.

    .. important

        The minus sign in the :math:`k_x(x, y, z)` function is needed to return
        the **upward** component of the acceleration. [Nagy2000]_ and
        [Nagy2002]_ equation corresponds to the *downward* coordinate.

    The :math:`\operatorname{safe-ln}` and :math:`\operatorname{safe-arctan}`
    functions are defined as follows:

    .. math::

        \operatorname{safe-ln}(x, r) =
        \begin{cases}
            0 & x = 0, r = 0 \\
            \ln(x + r) & x \ge 0 \\
            \ln((r^2 - x^2) / (r - x)) & x < 0, r \ne -x \\
            -\ln(-2 x) & x < 0, r = -x
        \end{cases}

    .. math::

        \operatorname{safe-arctan} \left( y, x \right) =
        \begin{cases}
            \text{arctan}\left( \frac{y}{x} \right) & x \ne 0 \\
            \frac{\pi}{2} & x = 0 \quad \text{and} \quad y > 0 \\
            -\frac{\pi}{2} & x = 0 \quad \text{and} \quad y < 0 \\
            0 & x = 0 \quad \text{and} \quad y = 0 \\
        \end{cases}

    These were defined after [Fukushima2020]_ and guarantee a good accuracy on
    any observation point.

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    return (
        GRAVITATIONAL_CONST
        * density
        * _evaluate_kernel(
            easting,
            northing,
            upward,
            prism_west,
            prism_east,
            prism_south,
            prism_north,
            prism_bottom,
            prism_top,
            kernel_u,
        )
    )


@jit(nopython=True)
def gravity_ee(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    density,
):
    r"""
    Easting-easting component of the gravitational tensor due to a prism

    Returns the easting-easting component of the gravitational tensor produced
    by a single rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    density : float
        Density of the rectangular prism in kilograms per cubic meter.

    Returns
    -------
    g_ee : float
        Easting-easting component of the gravitational tensor generated by the
        rectangular prism on the observation point in
        :math:`\text{m}/\text{s}^2`.
        Return ``numpy.nan`` if the observation point falls in a singular
        point: prism vertices or prism edges perpendicular to the easting
        direction.

    Notes
    -----
    Returns the easting-easting component :math:`g_{xx}(\mathbf{p})` of the
    gravitational tensor :math:`\mathbf{T}` on the observation point
    :math:`\mathbf{p} = (x_p, y_p, z_p)` generated by a single rectangular
    prism defined by its boundaries :math:`x_1, x_2, y_1, y_2, z_1, z_2` and
    with a density :math:`\rho`:

    .. math::

        g_{xx}(\mathbf{p}) =
            G
            \rho \,\,
            \Bigg\lvert \Bigg\lvert \Bigg\lvert
            k_{xx}(x, y, z)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}

    where

    .. math::

        k_{xx}(x, y, z) = - \operatorname{safe-arctan} \left( yz, xr \right),

    .. math::

        r = \sqrt{x^2 + y^2 + z^2},

    and

    .. math::

        X_1 = x_1 - x_p \\
        X_2 = x_2 - x_p \\
        Y_1 = y_1 - y_p \\
        Y_2 = y_2 - y_p \\
        Z_1 = z_1 - z_p \\
        Z_2 = z_2 - z_p

    are the shifted coordinates of the prism boundaries and :math:`G` is the
    Universal Gravitational Constant.

    The :math:`\operatorname{safe-arctan}` function is defined as follows:

    .. math::

        \operatorname{safe-arctan} \left( y, x \right) =
        \begin{cases}
            \text{arctan}\left( \frac{y}{x} \right) & x \ne 0 \\
            \frac{\pi}{2} & x = 0 \quad \text{and} \quad y > 0 \\
            -\frac{\pi}{2} & x = 0 \quad \text{and} \quad y < 0 \\
            0 & x = 0 \quad \text{and} \quad y = 0 \\
        \end{cases}

    It was defined after [Fukushima2020]_ and guarantee a good accuracy on any
    observation point.

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    # Return nan if the observation point falls on a singular point.
    # For the g_ee this are edges perpendicular to the easting direction
    # (parallel to northing and upward)
    if is_point_on_northing_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ) or is_point_on_upward_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return np.nan
    # Evaluate kernel function on each vertex of the prism
    result = _evaluate_kernel(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
        kernel_ee,
    )
    # Add 4 pi if computing on the eastern face to return the limit approaching
    # from outside (approaching from the east)
    if is_point_on_east_face(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        result += 4 * np.pi
    return GRAVITATIONAL_CONST * density * result


@jit(nopython=True)
def gravity_nn(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    density,
):
    r"""
    Northing-northing component of the gravitational tensor due to a prism

    Returns the northing-northing component of the gravitational tensor
    produced by a single rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    density : float
        Density of the rectangular prism in kilograms per cubic meter.

    Returns
    -------
    g_nn : float
        Northing-northing component of the gravitational tensor generated by
        the rectangular prism on the observation point in
        :math:`\text{m}/\text{s}^2`.
        Return ``numpy.nan`` if the observation point falls in a singular
        point: prism vertices or prism edges perpendicular to the northing
        direction.

    Notes
    -----
    Returns the northing-northing component :math:`g_{yy}(\mathbf{p})` of the
    gravitational tensor :math:`\mathbf{T}` on the observation point
    :math:`\mathbf{p} = (x_p, y_p, z_p)` generated by a single rectangular
    prism defined by its boundaries :math:`x_1, x_2, y_1, y_2, z_1, z_2` and
    with a density :math:`\rho`:

    .. math::

        g_{yy}(\mathbf{p}) =
            G
            \rho \,\,
            \Bigg\lvert \Bigg\lvert \Bigg\lvert
            k_{yy}(x, y, z)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}

    where

    .. math::

        k_{yy}(x, y, z) = - \operatorname{safe-arctan} \left( zx, yr \right),

    .. math::

        r = \sqrt{x^2 + y^2 + z^2},

    and

    .. math::

        X_1 = x_1 - x_p \\
        X_2 = x_2 - x_p \\
        Y_1 = y_1 - y_p \\
        Y_2 = y_2 - y_p \\
        Z_1 = z_1 - z_p \\
        Z_2 = z_2 - z_p

    are the shifted coordinates of the prism boundaries and :math:`G` is the
    Universal Gravitational Constant.

    The :math:`\operatorname{safe-arctan}` function is defined as follows:

    .. math::

        \operatorname{safe-arctan} \left( y, x \right) =
        \begin{cases}
            \text{arctan}\left( \frac{y}{x} \right) & x \ne 0 \\
            \frac{\pi}{2} & x = 0 \quad \text{and} \quad y > 0 \\
            -\frac{\pi}{2} & x = 0 \quad \text{and} \quad y < 0 \\
            0 & x = 0 \quad \text{and} \quad y = 0 \\
        \end{cases}

    It was defined after [Fukushima2020]_ and guarantee a good accuracy on any
    observation point.

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    # Return nan if the observation point falls on a singular point.
    # For the g_nn this are edges perpendicular to the northing direction
    # (parallel to easting and upward)
    if is_point_on_easting_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ) or is_point_on_upward_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return np.nan
    # Evaluate kernel function on each vertex of the prism
    result = _evaluate_kernel(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
        kernel_nn,
    )
    # Add 4 pi if computing on the northern face to return the limit
    # approaching from outside (approaching from the north)
    if is_point_on_north_face(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        result += 4 * np.pi
    return GRAVITATIONAL_CONST * density * result


@jit(nopython=True)
def gravity_uu(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    density,
):
    r"""
    Upward-upward component of the gravitational tensor due to a prism

    Returns the northing-northing component of the gravitational tensor
    produced by a single rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    density : float
        Density of the rectangular prism in kilograms per cubic meter.

    Returns
    -------
    g_uu : float
        Upward-upward component of the gravitational tensor generated by
        the rectangular prism on the observation point in
        :math:`\text{m}/\text{s}^2`.
        Return ``numpy.nan`` if the observation point falls in a singular
        point: prism vertices or prism edges perpendicular to the upward
        direction.

    Notes
    -----
    Returns the upward-upward component :math:`g_{zz}(\mathbf{p})` of the
    gravitational tensor :math:`\mathbf{T}` on the observation point
    :math:`\mathbf{p} = (x_p, y_p, z_p)` generated by a single rectangular
    prism defined by its boundaries :math:`x_1, x_2, y_1, y_2, z_1, z_2` and
    with a density :math:`\rho`:

    .. math::

        g_{zz}(\mathbf{p}) =
            G
            \rho \,\,
            \Bigg\lvert \Bigg\lvert \Bigg\lvert
            k_{zz}(x, y, z)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}

    where

    .. math::

        k_{zz}(x, y, z) = - \operatorname{safe-arctan} \left( xy, zr \right),

    .. math::

        r = \sqrt{x^2 + y^2 + z^2},

    and

    .. math::

        X_1 = x_1 - x_p \\
        X_2 = x_2 - x_p \\
        Y_1 = y_1 - y_p \\
        Y_2 = y_2 - y_p \\
        Z_1 = z_1 - z_p \\
        Z_2 = z_2 - z_p

    are the shifted coordinates of the prism boundaries and :math:`G` is the
    Universal Gravitational Constant.

    The :math:`\operatorname{safe-arctan}` function is defined as follows:

    .. math::

        \operatorname{safe-arctan} \left( y, x \right) =
        \begin{cases}
            \text{arctan}\left( \frac{y}{x} \right) & x \ne 0 \\
            \frac{\pi}{2} & x = 0 \quad \text{and} \quad y > 0 \\
            -\frac{\pi}{2} & x = 0 \quad \text{and} \quad y < 0 \\
            0 & x = 0 \quad \text{and} \quad y = 0 \\
        \end{cases}

    It was defined after [Fukushima2020]_ and guarantee a good accuracy on any
    observation point.

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    # Return nan if the observation point falls on a singular point.
    # For the g_uu this are edges perpendicular to the upward direction
    # (parallel to easting and northing)
    if is_point_on_easting_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ) or is_point_on_northing_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return np.nan
    # Evaluate kernel function on each vertex of the prism
    result = _evaluate_kernel(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
        kernel_uu,
    )
    # Add 4 pi if computing on the top face to return the limit approaching
    # from outside (approaching from the top)
    if is_point_on_top_face(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        result += 4 * np.pi
    return GRAVITATIONAL_CONST * density * result


@jit(nopython=True)
def gravity_en(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    density,
):
    r"""
    Easting-northing component of the gravitational tensor due to a prism

    Returns the northing-northing component of the gravitational tensor
    produced by a single rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    density : float
        Density of the rectangular prism in kilograms per cubic meter.

    Returns
    -------
    g_en : float
        Easting-northing component of the gravitational tensor generated by the
        rectangular prism on the observation point in
        :math:`\text{m}/\text{s}^2`.
        Return ``numpy.nan`` if the observation point falls in a singular
        point: prism vertices or prism edges parallel to the upward direction.

    Notes
    -----
    Returns the easting-northing component :math:`g_{xy}(\mathbf{p})` of the
    gravitational tensor :math:`\mathbf{T}` on the observation point
    :math:`\mathbf{p} = (x_p, y_p, z_p)` generated by a single rectangular
    prism defined by its boundaries :math:`x_1, x_2, y_1, y_2, z_1, z_2` and
    with a density :math:`\rho`:

    .. math::

        g_{xy}(\mathbf{p}) =
            G
            \rho \,\,
            \Bigg\lvert \Bigg\lvert \Bigg\lvert
            k_{xy}(x, y, z)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}

    where

    .. math::

        k_{xy}(x, y, z) = \operatorname{safe-ln} \left( z, r \right),

    .. math::

        r = \sqrt{x^2 + y^2 + z^2},

    and

    .. math::

        X_1 = x_1 - x_p \\
        X_2 = x_2 - x_p \\
        Y_1 = y_1 - y_p \\
        Y_2 = y_2 - y_p \\
        Z_1 = z_1 - z_p \\
        Z_2 = z_2 - z_p

    are the shifted coordinates of the prism boundaries and :math:`G` is the
    Universal Gravitational Constant.

    The :math:`\operatorname{safe-ln}` function is defined as follows:

    .. math::

        \operatorname{safe-ln}(x, r) =
        \begin{cases}
            0 & x = 0, r = 0 \\
            \ln(x + r) & x \ge 0 \\
            \ln((r^2 - x^2) / (r - x)) & x < 0, r \ne -x \\
            -\ln(-2 x) & x < 0, r = -x
        \end{cases}

    It was defined after [Fukushima2020]_ and guarantee a good accuracy on any
    observation point.

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    # Return nan if the observation point falls on a singular point.
    # For g_en this are edges parallel to to the upward direction
    if is_point_on_upward_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return np.nan
    return (
        GRAVITATIONAL_CONST
        * density
        * _evaluate_kernel(
            easting,
            northing,
            upward,
            prism_west,
            prism_east,
            prism_south,
            prism_north,
            prism_bottom,
            prism_top,
            kernel_en,
        )
    )


@jit(nopython=True)
def gravity_eu(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    density,
):
    r"""
    Easting-upward component of the gravitational tensor due to a prism

    Returns the easting-upward component of the gravitational tensor
    produced by a single rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    density : float
        Density of the rectangular prism in kilograms per cubic meter.

    Returns
    -------
    g_eu : float
        Easting-upward component of the gravitational tensor generated by the
        rectangular prism on the observation point in
        :math:`\text{m}/\text{s}^2`.
        Return ``numpy.nan`` if the observation point falls in a singular
        point: prism vertices or prism edges parallel to the northing
        direction.

    Notes
    -----
    Returns the easting-upward component :math:`g_{xz}(\mathbf{p})` of the
    gravitational tensor :math:`\mathbf{T}` on the observation point
    :math:`\mathbf{p} = (x_p, y_p, z_p)` generated by a single rectangular
    prism defined by its boundaries :math:`x_1, x_2, y_1, y_2, z_1, z_2` and
    with a density :math:`\rho`:

    .. math::

        g_{xz}(\mathbf{p}) =
            G
            \rho \,\,
            \Bigg\lvert \Bigg\lvert \Bigg\lvert
            k_{xz}(x, y, z)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}

    where

    .. math::

        k_{xz}(x, y, z) = \operatorname{safe-ln} \left( y, r \right),

    .. math::

        r = \sqrt{x^2 + y^2 + z^2},

    and

    .. math::

        X_1 = x_1 - x_p \\
        X_2 = x_2 - x_p \\
        Y_1 = y_1 - y_p \\
        Y_2 = y_2 - y_p \\
        Z_1 = z_1 - z_p \\
        Z_2 = z_2 - z_p

    are the shifted coordinates of the prism boundaries and :math:`G` is the
    Universal Gravitational Constant.

    The :math:`\operatorname{safe-ln}` function is defined as follows:

    .. math::

        \operatorname{safe-ln}(x, r) =
        \begin{cases}
            0 & x = 0, r = 0 \\
            \ln(x + r) & x \ge 0 \\
            \ln((r^2 - x^2) / (r - x)) & x < 0, r \ne -x \\
            -\ln(-2 x) & x < 0, r = -x
        \end{cases}

    It was defined after [Fukushima2020]_ and guarantee a good accuracy on any
    observation point.

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    # Return nan if the observation point falls on a singular point.
    # For g_eu this are edges parallel to to the northing direction
    if is_point_on_northing_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return np.nan
    return (
        GRAVITATIONAL_CONST
        * density
        * _evaluate_kernel(
            easting,
            northing,
            upward,
            prism_west,
            prism_east,
            prism_south,
            prism_north,
            prism_bottom,
            prism_top,
            kernel_eu,
        )
    )


@jit(nopython=True)
def gravity_nu(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    density,
):
    r"""
    Northing-upward component of the gravitational tensor due to a prism

    Returns the northing-upward component of the gravitational tensor
    produced by a single rectangular prism on a single computation point.

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    density : float
        Density of the rectangular prism in kilograms per cubic meter.

    Returns
    -------
    g_nu : float
        Northing-upward component of the gravitational tensor generated by the
        rectangular prism on the observation point in
        :math:`\text{m}/\text{s}^2`.
        Return ``numpy.nan`` if the observation point falls in a singular
        point: prism vertices or prism edges parallel to the easting
        direction.

    Notes
    -----
    Returns the northing-upward component :math:`g_{yz}(\mathbf{p})` of the
    gravitational tensor :math:`\mathbf{T}` on the observation point
    :math:`\mathbf{p} = (x_p, y_p, z_p)` generated by a single rectangular
    prism defined by its boundaries :math:`x_1, x_2, y_1, y_2, z_1, z_2` and
    with a density :math:`\rho`:

    .. math::

        g_{yz}(\mathbf{p}) =
            G
            \rho \,\,
            \Bigg\lvert \Bigg\lvert \Bigg\lvert
            k_{yz}(x, y, z)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}

    where

    .. math::

        k_{yz}(x, y, z) = \operatorname{safe-ln} \left( x, r \right),

    .. math::

        r = \sqrt{x^2 + y^2 + z^2},

    and

    .. math::

        X_1 = x_1 - x_p \\
        X_2 = x_2 - x_p \\
        Y_1 = y_1 - y_p \\
        Y_2 = y_2 - y_p \\
        Z_1 = z_1 - z_p \\
        Z_2 = z_2 - z_p

    are the shifted coordinates of the prism boundaries and :math:`G` is the
    Universal Gravitational Constant.

    The :math:`\operatorname{safe-ln}` function is defined as follows:

    .. math::

        \operatorname{safe-ln}(x, r) =
        \begin{cases}
            0 & x = 0, r = 0 \\
            \ln(x + r) & x \ge 0 \\
            \ln((r^2 - x^2) / (r - x)) & x < 0, r \ne -x \\
            -\ln(-2 x) & x < 0, r = -x
        \end{cases}

    It was defined after [Fukushima2020]_ and guarantee a good accuracy on any
    observation point.

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    - [Fukushima2020]_
    """
    # Return nan if the observation point falls on a singular point.
    # For g_nu this are edges parallel to to the easting direction
    if is_point_on_easting_edge(
        easting,
        northing,
        upward,
        prism_west,
        prism_east,
        prism_south,
        prism_north,
        prism_bottom,
        prism_top,
    ):
        return np.nan
    return (
        GRAVITATIONAL_CONST
        * density
        * _evaluate_kernel(
            easting,
            northing,
            upward,
            prism_west,
            prism_east,
            prism_south,
            prism_north,
            prism_bottom,
            prism_top,
            kernel_nu,
        )
    )


@jit(nopython=True)
def _evaluate_kernel(
    easting,
    northing,
    upward,
    prism_west,
    prism_east,
    prism_south,
    prism_north,
    prism_bottom,
    prism_top,
    kernel,
):
    r"""
    Evaluate a kernel function on every shifted vertex of a prism

    Parameters
    ----------
    easting, northing, upward : float
        Easting, northing and upward coordinates of the observation point. Must
        be in meters.
    prism_west, prism_east, prism_south, prism_north, prism_bottom, prism_top : float
        The boundaries of the prism. Must be in meters.
    kernel : callable
        Kernel function that will be evaluated on each one of the shifted
        vertices of the prism.

    Returns
    -------
    result : float
        Evaluation of the kernel function on each one of the vertices of the
        prism.

    Notes
    -----
    This function evaluates a numerical kernel :math:`k(x, y, z)` on each one
    of the vertices of the prism:

    .. math::

        v(\mathbf{p}) =
            \Bigg\lvert \Bigg\lvert \Bigg\lvert
            k(x, y, z)
            \Bigg\rvert_{X_1}^{X_2}
            \Bigg\rvert_{Y_1}^{Y_2}
            \Bigg\rvert_{Z_1}^{Z_2}

    where :math:`X_1`, :math:`X_2`, :math:`Y_1`, :math:`Y_2`, :math:`Z_1` and
    :math:`Z_2` are boundaries of the rectangular prism in the *shifted
    coordinates* defined by the Cartesian coordinate system with its origin
    located on the observation point :math:`\mathbf{p}`.

    References
    ----------
    - [Nagy2000]_
    - [Nagy2002]_
    """
    # Initialize result float to zero
    result = 0
    # Iterate over the vertices of the prism
    for i in range(2):
        # Compute shifted easting coordinate
        if i == 0:
            shift_east = prism_east - easting
        else:
            shift_east = prism_west - easting
        shift_east_sq = shift_east**2
        for j in range(2):
            # Compute shifted northing coordinate
            if j == 0:
                shift_north = prism_north - northing
            else:
                shift_north = prism_south - northing
            shift_north_sq = shift_north**2
            for k in range(2):
                # Compute shifted upward coordinate
                if k == 0:
                    shift_upward = prism_top - upward
                else:
                    shift_upward = prism_bottom - upward
                shift_upward_sq = shift_upward**2
                # Compute the radius
                radius = np.sqrt(shift_east_sq + shift_north_sq + shift_upward_sq)
                # If i, j or k is 1, the corresponding shifted
                # coordinate will refer to the lower boundary,
                # meaning the corresponding term should have a minus
                # sign.
                result += (-1) ** (i + j + k) * kernel(
                    shift_east, shift_north, shift_upward, radius
                )
    return result



# Copyright (c) 2022 The Choclo Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
# Import functions/classes to make the public API
from . import dipole, point, prism
from ._version import __version__

# Append a leading "v" to the generated version by setuptools_scm
__version__ = f"v{__version__}"



# Copyright (c) 2022 The Choclo Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Additional tests to prism kernels.
"""

import numpy as np
import pytest

from choclo.prism import (
    kernel_ee,
    kernel_eee,
    kernel_een,
    kernel_eeu,
    kernel_en,
    kernel_enn,
    kernel_enu,
    kernel_eu,
    kernel_euu,
    kernel_nn,
    kernel_nnn,
    kernel_nnu,
    kernel_nu,
    kernel_nuu,
    kernel_uu,
    kernel_uuu,
)


class TestThirdOrderKernelsOnVertex:
    """
    Test if third-order kernels evaluated on a vertex return np.nan.

    This functionality is meant to be supported for the public usage of this
    kernels, although the forward modelling functions within Choclo won't use
    it since they filter out singular points before they call kernels.
    """

    SECOND_ORDER_KERNELS = (
        kernel_ee,
        kernel_nn,
        kernel_uu,
        kernel_en,
        kernel_eu,
        kernel_nu,
    )
    THIRD_ORDER_KERNELS = (
        kernel_eee,
        kernel_nnn,
        kernel_uuu,
        kernel_een,
        kernel_eeu,
        kernel_enn,
        kernel_nnu,
        kernel_euu,
        kernel_nuu,
        kernel_enu,
    )

    @pytest.mark.parametrize("kernel", THIRD_ORDER_KERNELS)
    def test_third_order_kernels_on_vertex(self, kernel):
        easting, northing, upward = 0.0, 0.0, 0.0
        radius = 0.0
        result = kernel(easting, northing, upward, radius)
        assert np.isnan(result)

    @pytest.mark.parametrize("kernel", SECOND_ORDER_KERNELS)
    def test_second_order_kernels_on_vertex(self, kernel):
        easting, northing, upward = 0.0, 0.0, 0.0
        radius = 0.0
        result = kernel(easting, northing, upward, radius)
        assert np.isnan(result)



# Copyright (c) 2022 The Choclo Developers.
# Distributed under the terms of the BSD 3-Clause License.
# SPDX-License-Identifier: BSD-3-Clause
#
# This code is part of the Fatiando a Terra project (https://www.fatiando.org)
#
"""
Test magnetic forward modelling functions for a dipole
"""
import numpy as np
import numpy.testing as npt
import pytest

from ..dipole import magnetic_e, magnetic_field, magnetic_n, magnetic_u


@pytest.fixture(name="sample_dipole")
def fixture_sample_dipole():
    """
    Return the location of a sample dipole
    """
    return 40.5, 32.4, -15.3


@pytest.fixture(name="sample_magnetic_moment")
def fixture_sample_magnetic_moment():
    """
    Return a sample magnetic moment
    """
    return np.array([780.3, -230.4, 1030])


@pytest.fixture(name="sample_3d_grid")
def fixture_sample_3d_grid(sample_dipole):
    """
    Return a set of observation points in a 3d grid

    These observation points are centered around the dipole, but they don't
    include its position.
    """
    easting = np.linspace(-40, 40, 11)
    northing = np.linspace(-40, 40, 11)
    upward = np.linspace(-40, 40, 11)
    # Compute meshgrid
    easting, northing, upward = np.meshgrid(easting, northing, upward)
    # Remove the location of the dipole
    is_dipole = (easting == 0) & (northing == 0) & (upward == 0)
    easting = easting[np.logical_not(is_dipole)]
    northing = northing[np.logical_not(is_dipole)]
    upward = upward[np.logical_not(is_dipole)]
    # Shift coordinates
    easting += sample_dipole[0]
    northing += sample_dipole[1]
    upward += sample_dipole[2]
    return easting, northing, upward


class TestSymmetryBe:
    """
    Test symmetry of easting component of the magnetic field
    """

    atol = 1e-22  # absolute tolerance for values near zero

    @pytest.mark.parametrize(
        "magnetic_moment",
        [
            (500, 0, 0),
            (-500, 0, 0),
            (0, 500, 0),
            (0, -500, 0),
            (0, 0, 500),
            (0, 0, -500),
        ],
    )
    def test_symmetry_across_easting_northing(
        self, sample_3d_grid, sample_dipole, magnetic_moment
    ):
        """
        Test symmetry of magnetic_e across the easting-northing plane that
        passes through the location of the dipole
        """
        easting, northing, upward = sample_3d_grid
        # Keep only the observation points that are above the dipole
        is_top = upward > sample_dipole[2]
        easting = easting[is_top]
        northing = northing[is_top]
        upward_top = upward[is_top]
        # Create a symmetrical upward coordinate for points below the dipole
        upward_bottom = 2 * sample_dipole[2] - upward_top
        # Compute magnetic_e on every observation point
        magnetic_moment = np.array(magnetic_moment)
        b_e_top = np.array(
            list(
                magnetic_e(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting.ravel(), northing.ravel(), upward_top.ravel()
                )
            )
        )
        b_e_bottom = np.array(
            list(
                magnetic_e(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting.ravel(), northing.ravel(), upward_bottom.ravel()
                )
            )
        )
        # Check symmetry between top and bottom
        if magnetic_moment[2] != 0:
            npt.assert_allclose(b_e_top, -b_e_bottom, atol=self.atol)
        else:
            npt.assert_allclose(b_e_top, b_e_bottom, atol=self.atol)

    @pytest.mark.parametrize(
        "magnetic_moment",
        [
            (500, 0, 0),
            (-500, 0, 0),
            (0, 500, 0),
            (0, -500, 0),
            (0, 0, 500),
            (0, 0, -500),
        ],
    )
    def test_symmetry_across_easting_upward(
        self, sample_3d_grid, sample_dipole, magnetic_moment
    ):
        """
        Test symmetry of magnetic_e across the easting-upward plane that
        passes through the location of the dipole
        """
        easting, northing, upward = sample_3d_grid
        # Keep only the observation points that are north the dipole
        is_north = northing > sample_dipole[1]
        easting = easting[is_north]
        northing_north = northing[is_north]
        upward = upward[is_north]
        # Create a symmetrical upward coordinate for points south the dipole
        northing_south = 2 * sample_dipole[1] - northing_north
        # Compute magnetic_e on every observation point
        magnetic_moment = np.array(magnetic_moment)
        b_e_north = np.array(
            list(
                magnetic_e(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting.ravel(), northing_north.ravel(), upward.ravel()
                )
            )
        )
        b_e_south = np.array(
            list(
                magnetic_e(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting.ravel(), northing_south.ravel(), upward.ravel()
                )
            )
        )
        # Check symmetry between south and north
        if magnetic_moment[1] != 0:
            npt.assert_allclose(b_e_north, -b_e_south, atol=self.atol)
        else:
            npt.assert_allclose(b_e_north, b_e_south, atol=self.atol)

    @pytest.mark.parametrize(
        "magnetic_moment",
        [
            (500, 0, 0),
            (-500, 0, 0),
            (0, 500, 0),
            (0, -500, 0),
            (0, 0, 500),
            (0, 0, -500),
        ],
    )
    def test_symmetry_across_northing_upward(
        self, sample_3d_grid, sample_dipole, magnetic_moment
    ):
        """
        Test symmetry of magnetic_e across the northing-upward plane that
        passes through the location of the dipole
        """
        easting, northing, upward = sample_3d_grid
        # Keep only the observation points that are east the dipole
        is_east = easting > sample_dipole[0]
        easting_east = easting[is_east]
        northing = northing[is_east]
        upward = upward[is_east]
        # Create a symmetrical upward coordinate for points west the dipole
        easting_west = 2 * sample_dipole[0] - easting_east
        # Compute magnetic_e on every observation point
        magnetic_moment = np.array(magnetic_moment)
        b_e_east = np.array(
            list(
                magnetic_e(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting_east.ravel(), northing.ravel(), upward.ravel()
                )
            )
        )
        b_e_west = np.array(
            list(
                magnetic_e(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting_west.ravel(), northing.ravel(), upward.ravel()
                )
            )
        )
        # Check symmetry between west and east
        if magnetic_moment[0] != 0:
            npt.assert_allclose(b_e_east, b_e_west, atol=self.atol)
        else:
            npt.assert_allclose(b_e_east, -b_e_west, atol=self.atol)

    def test_symmetry_when_flipping(self, sample_3d_grid, sample_dipole):
        """
        Test symmetry of magnetic_e when flipping its direction
        """
        easting, northing, upward = sample_3d_grid
        # Keep only points with easting on east of the dipole
        is_east_or_equal = easting >= sample_dipole[0]
        easting = easting[is_east_or_equal]
        northing = northing[is_east_or_equal]
        upward = upward[is_east_or_equal]
        # Define two magnetic moments
        magnetic_moment_east = np.array([500.0, 0, 0])
        magnetic_moment_west = np.array([-500.0, 0, 0])
        # Compute the magnetic field generated by each moment on the
        # observation points
        b_e_east = np.array(
            list(
                magnetic_e(e, n, u, *sample_dipole, *magnetic_moment_east)
                for e, n, u in zip(easting, northing, upward)
            )
        )
        b_e_west = np.array(
            list(
                magnetic_e(e, n, u, *sample_dipole, *magnetic_moment_west)
                for e, n, u in zip(easting, northing, upward)
            )
        )
        # Check if the sign gets inverted
        npt.assert_allclose(b_e_east, -b_e_west)


class TestSymmetryBn:
    """
    Test symmetry of easting component of the magnetic field
    """

    atol = 1e-22  # absolute tolerance for values near zero

    @pytest.mark.parametrize(
        "magnetic_moment",
        [
            (500, 0, 0),
            (-500, 0, 0),
            (0, 500, 0),
            (0, -500, 0),
            (0, 0, 500),
            (0, 0, -500),
        ],
    )
    def test_symmetry_across_easting_northing(
        self, sample_3d_grid, sample_dipole, magnetic_moment
    ):
        """
        Test symmetry of magnetic_n across the easting-northing plane that
        passes through the location of the dipole
        """
        easting, northing, upward = sample_3d_grid
        # Keep only the observation points that are above the dipole
        is_top = upward > sample_dipole[2]
        easting = easting[is_top]
        northing = northing[is_top]
        upward_top = upward[is_top]
        # Create a symmetrical upward coordinate for points below the dipole
        upward_bottom = 2 * sample_dipole[2] - upward_top
        # Compute magnetic_n on every observation point
        magnetic_moment = np.array(magnetic_moment)
        b_n_top = np.array(
            list(
                magnetic_n(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting.ravel(), northing.ravel(), upward_top.ravel()
                )
            )
        )
        b_n_bottom = np.array(
            list(
                magnetic_n(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting.ravel(), northing.ravel(), upward_bottom.ravel()
                )
            )
        )
        # Check symmetry between top and bottom
        if magnetic_moment[2] != 0:
            npt.assert_allclose(b_n_top, -b_n_bottom, atol=self.atol)
        else:
            npt.assert_allclose(b_n_top, b_n_bottom, atol=self.atol)

    @pytest.mark.parametrize(
        "magnetic_moment",
        [
            (500, 0, 0),
            (-500, 0, 0),
            (0, 500, 0),
            (0, -500, 0),
            (0, 0, 500),
            (0, 0, -500),
        ],
    )
    def test_symmetry_across_easting_upward(
        self, sample_3d_grid, sample_dipole, magnetic_moment
    ):
        """
        Test symmetry of magnetic_n across the easting-upward plane that
        passes through the location of the dipole
        """
        easting, northing, upward = sample_3d_grid
        # Keep only the observation points that are north the dipole
        is_north = northing > sample_dipole[1]
        easting = easting[is_north]
        northing_north = northing[is_north]
        upward = upward[is_north]
        # Create a symmetrical upward coordinate for points south the dipole
        northing_south = 2 * sample_dipole[1] - northing_north
        # Compute magnetic_n on every observation point
        magnetic_moment = np.array(magnetic_moment)
        b_n_north = np.array(
            list(
                magnetic_n(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting.ravel(), northing_north.ravel(), upward.ravel()
                )
            )
        )
        b_n_south = np.array(
            list(
                magnetic_n(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting.ravel(), northing_south.ravel(), upward.ravel()
                )
            )
        )
        # Check symmetry between south and north
        if magnetic_moment[1] != 0:
            npt.assert_allclose(b_n_north, b_n_south, atol=self.atol)
        else:
            npt.assert_allclose(b_n_north, -b_n_south, atol=self.atol)

    @pytest.mark.parametrize(
        "magnetic_moment",
        [
            (500, 0, 0),
            (-500, 0, 0),
            (0, 500, 0),
            (0, -500, 0),
            (0, 0, 500),
            (0, 0, -500),
        ],
    )
    def test_symmetry_across_northing_upward(
        self, sample_3d_grid, sample_dipole, magnetic_moment
    ):
        """
        Test symmetry of magnetic_n across the northing-upward plane that
        passes through the location of the dipole
        """
        easting, northing, upward = sample_3d_grid
        # Keep only the observation points that are east the dipole
        is_east = easting > sample_dipole[0]
        easting_east = easting[is_east]
        northing = northing[is_east]
        upward = upward[is_east]
        # Create a symmetrical upward coordinate for points west the dipole
        easting_west = 2 * sample_dipole[0] - easting_east
        # Compute magnetic_n on every observation point
        magnetic_moment = np.array(magnetic_moment)
        b_n_east = np.array(
            list(
                magnetic_n(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting_east.ravel(), northing.ravel(), upward.ravel()
                )
            )
        )
        b_n_west = np.array(
            list(
                magnetic_n(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting_west.ravel(), northing.ravel(), upward.ravel()
                )
            )
        )
        # Check symmetry between west and east
        if magnetic_moment[0] != 0:
            npt.assert_allclose(b_n_east, -b_n_west, atol=self.atol)
        else:
            npt.assert_allclose(b_n_east, b_n_west, atol=self.atol)

    def test_symmetry_when_flipping(self, sample_3d_grid, sample_dipole):
        """
        Test symmetry of magnetic_n when flipping its direction
        """
        easting, northing, upward = sample_3d_grid
        # Keep only points with northing on north of the dipole
        is_north_or_equal = northing >= sample_dipole[1]
        easting = easting[is_north_or_equal]
        northing = northing[is_north_or_equal]
        upward = upward[is_north_or_equal]
        # Define two magnetic moments
        magnetic_moment_north = np.array([0, 500.0, 0])
        magnetic_moment_south = np.array([0, -500.0, 0])
        # Compute the magnetic field generated by each moment on the
        # observation points
        b_n_north = np.array(
            list(
                magnetic_n(e, n, u, *sample_dipole, *magnetic_moment_north)
                for e, n, u in zip(easting, northing, upward)
            )
        )
        b_n_south = np.array(
            list(
                magnetic_n(e, n, u, *sample_dipole, *magnetic_moment_south)
                for e, n, u in zip(easting, northing, upward)
            )
        )
        # Check if the sign gets inverted
        npt.assert_allclose(b_n_north, -b_n_south)


class TestSymmetryBu:
    """
    Test symmetry of upward component of the magnetic field
    """

    atol = 1e-22  # absolute tolerance for values near zero

    @pytest.mark.parametrize(
        "magnetic_moment",
        [
            (500, 0, 0),
            (-500, 0, 0),
            (0, 500, 0),
            (0, -500, 0),
            (0, 0, 500),
            (0, 0, -500),
        ],
    )
    def test_symmetry_across_easting_northing(
        self, sample_3d_grid, sample_dipole, magnetic_moment
    ):
        """
        Test symmetry of magnetic_e across the easting-northing plane that
        passes through the location of the dipole
        """
        easting, northing, upward = sample_3d_grid
        # Keep only the observation points that are above the dipole
        is_top = upward > sample_dipole[2]
        easting = easting[is_top]
        northing = northing[is_top]
        upward_top = upward[is_top]
        # Create a symmetrical upward coordinate for points below the dipole
        upward_bottom = 2 * sample_dipole[2] - upward_top
        # Compute magnetic_u on every observation point
        magnetic_moment = np.array(magnetic_moment)
        b_u_top = np.array(
            list(
                magnetic_u(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting.ravel(), northing.ravel(), upward_top.ravel()
                )
            )
        )
        b_u_bottom = np.array(
            list(
                magnetic_u(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting.ravel(), northing.ravel(), upward_bottom.ravel()
                )
            )
        )
        # Check symmetry between top and bottom
        if magnetic_moment[2] != 0:
            npt.assert_allclose(b_u_top, b_u_bottom, atol=self.atol)
        else:
            npt.assert_allclose(b_u_top, -b_u_bottom, atol=self.atol)

    @pytest.mark.parametrize(
        "magnetic_moment",
        [
            (500, 0, 0),
            (-500, 0, 0),
            (0, 500, 0),
            (0, -500, 0),
            (0, 0, 500),
            (0, 0, -500),
        ],
    )
    def test_symmetry_across_easting_upward(
        self, sample_3d_grid, sample_dipole, magnetic_moment
    ):
        """
        Test symmetry of magnetic_e across the easting-upward plane that
        passes through the location of the dipole
        """
        easting, northing, upward = sample_3d_grid
        # Keep only the observation points that are north the dipole
        is_north = northing > sample_dipole[1]
        easting = easting[is_north]
        northing_north = northing[is_north]
        upward = upward[is_north]
        # Create a symmetrical upward coordinate for points south the dipole
        northing_south = 2 * sample_dipole[1] - northing_north
        # Compute magnetic_u on every observation point
        magnetic_moment = np.array(magnetic_moment)
        b_u_north = np.array(
            list(
                magnetic_u(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting.ravel(), northing_north.ravel(), upward.ravel()
                )
            )
        )
        b_u_south = np.array(
            list(
                magnetic_u(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting.ravel(), northing_south.ravel(), upward.ravel()
                )
            )
        )
        # Check symmetry between south and north
        if magnetic_moment[1] != 0:
            npt.assert_allclose(b_u_north, -b_u_south, atol=self.atol)
        else:
            npt.assert_allclose(b_u_north, b_u_south, atol=self.atol)

    @pytest.mark.parametrize(
        "magnetic_moment",
        [
            (500, 0, 0),
            (-500, 0, 0),
            (0, 500, 0),
            (0, -500, 0),
            (0, 0, 500),
            (0, 0, -500),
        ],
    )
    def test_symmetry_across_northing_upward(
        self, sample_3d_grid, sample_dipole, magnetic_moment
    ):
        """
        Test symmetry of magnetic_e across the northing-upward plane that
        passes through the location of the dipole
        """
        easting, northing, upward = sample_3d_grid
        # Keep only the observation points that are east the dipole
        is_east = easting > sample_dipole[0]
        easting_east = easting[is_east]
        northing = northing[is_east]
        upward = upward[is_east]
        # Create a symmetrical upward coordinate for points west the dipole
        easting_west = 2 * sample_dipole[0] - easting_east
        # Compute magnetic_u on every observation point
        magnetic_moment = np.array(magnetic_moment)
        b_u_east = np.array(
            list(
                magnetic_u(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting_east.ravel(), northing.ravel(), upward.ravel()
                )
            )
        )
        b_u_west = np.array(
            list(
                magnetic_u(e, n, u, *sample_dipole, *magnetic_moment)
                for e, n, u in zip(
                    easting_west.ravel(), northing.ravel(), upward.ravel()
                )
            )
        )
        # Check symmetry between east and west
        if magnetic_moment[0] != 0:
            npt.assert_allclose(b_u_east, -b_u_west, atol=self.atol)
        else:
            npt.assert_allclose(b_u_east, b_u_west, atol=self.atol)

    def test_symmetry_when_flipping(self, sample_3d_grid, sample_dipole):
        """
        Test symmetry of magnetic_e when flipping its direction
        """
        easting, northing, upward = sample_3d_grid
        # Keep only points with upward on top of the dipole
        is_top_or_equal = upward >= sample_dipole[2]
        easting = easting[is_top_or_equal]
        northing = northing[is_top_or_equal]
        upward = upward[is_top_or_equal]
        # Define two magnetic moments
        magnetic_moment_up = np.array([0, 0, 500.0])
        magnetic_moment_down = np.array([0, 0, -500.0])
        # Compute the magnetic field generated by each moment on the
        # observation points
        b_u_up = np.array(
            list(
                magnetic_u(e, n, u, *sample_dipole, *magnetic_moment_up)
                for e, n, u in zip(easting, northing, upward)
            )
        )
        b_u_down = np.array(
            list(
                magnetic_u(e, n, u, *sample_dipole, *magnetic_moment_down)
                for e, n, u in zip(easting, northing, upward)
            )
        )
        # Check if the sign gets inverted
        npt.assert_allclose(b_u_up, -b_u_down)


class TestMagneticField:
    """
    Test magnetic_field against magnetic_easting, magnetic_northing and
    magnetic_upward

    Check if the components returned by magnetic_field match the individual
    ones computed by each one of the other functions.
    """

    def test_magnetic_field(
        self, sample_3d_grid, sample_dipole, sample_magnetic_moment
    ):
        """
        Test magnetic_field against each one of the other functions
        """
        # Compute all components of B using magnetic_field
        b = np.array(
            list(
                magnetic_field(e, n, u, *sample_dipole, *sample_magnetic_moment)
                for e, n, u in zip(*sample_3d_grid)
            )
        )
        b_e, b_n, b_u = tuple(b[:, i] for i in range(3))
        # Computed the individual fields
        b_e_expected = np.array(
            list(
                magnetic_e(e, n, u, *sample_dipole, *sample_magnetic_moment)
                for e, n, u in zip(*sample_3d_grid)
            )
        )
        b_n_expected = np.array(
            list(
                magnetic_n(e, n, u, *sample_dipole, *sample_magnetic_moment)
                for e, n, u in zip(*sample_3d_grid)
            )
        )
        b_u_expected = np.array(
            list(
                magnetic_u(e, n, u, *sample_dipole, *sample_magnetic_moment)
                for e, n, u in zip(*sample_3d_grid)
            )
        )
        npt.assert_allclose(b_e, b_e_expected)
        npt.assert_allclose(b_n, b_n_expected)
        npt.assert_allclose(b_u, b_u_expected)


class TestDivergenceOfB:
    """
    Test if the divergence of the magnetic field is equal to zero

    Compute the derivatives of B through finite differences
    """

    # Displacement used in the finite differences
    delta = 1e-5

    def get_b_ee_finite_differences(self, coordinates, dipole, magnetic_moment):
        """
        Compute b_ee using finite differences
        """
        # Get original coordinates
        easting, northing, upward = coordinates
        # Shift coordinates using delta
        easting_shifted = easting + self.delta
        # Compute b_e on original and shifted coordinates
        b_e = np.array(
            list(
                magnetic_e(e, n, u, *dipole, *magnetic_moment)
                for e, n, u in zip(easting, northing, upward)
            )
        )
        b_e_shifted = np.array(
            list(
                magnetic_e(e, n, u, *dipole, *magnetic_moment)
                for e, n, u in zip(easting_shifted, northing, upward)
            )
        )
        # Compute b_ee
        b_ee = (b_e_shifted - b_e) / self.delta
        return b_ee

    def get_b_nn_finite_differences(self, coordinates, dipole, magnetic_moment):
        """
        Compute b_nn using finite differences
        """
        # Get original coordinates
        easting, northing, upward = coordinates
        # Shift coordinates using delta
        northing_shifted = northing + self.delta
        # Compute b_e on original and shifted coordinates
        b_n = np.array(
            list(
                magnetic_n(e, n, u, *dipole, *magnetic_moment)
                for e, n, u in zip(easting, northing, upward)
            )
        )
        b_n_shifted = np.array(
            list(
                magnetic_n(e, n, u, *dipole, *magnetic_moment)
                for e, n, u in zip(easting, northing_shifted, upward)
            )
        )
        # Compute b_nn
        b_nn = (b_n_shifted - b_n) / self.delta
        return b_nn

    def get_b_uu_finite_differences(self, coordinates, dipole, magnetic_moment):
        """
        Compute b_uu using finite differences
        """
        # Get original coordinates
        easting, northing, upward = coordinates
        # Shift coordinates using delta
        upward_shifted = upward + self.delta
        # Compute b_e on original and shifted coordinates
        b_u = np.array(
            list(
                magnetic_u(e, n, u, *dipole, *magnetic_moment)
                for e, n, u in zip(easting, northing, upward)
            )
        )
        b_u_shifted = np.array(
            list(
                magnetic_u(e, n, u, *dipole, *magnetic_moment)
                for e, n, u in zip(easting, northing, upward_shifted)
            )
        )
        # Compute b_uu
        b_uu = (b_u_shifted - b_u) / self.delta
        return b_uu

    def test_divergence_of_b(
        self, sample_3d_grid, sample_dipole, sample_magnetic_moment
    ):
        # Compute b_ee, b_nn and b_uu using finite differences
        b_ee = self.get_b_ee_finite_differences(
            sample_3d_grid, sample_dipole, sample_magnetic_moment
        )
        b_nn = self.get_b_nn_finite_differences(
            sample_3d_grid, sample_dipole, sample_magnetic_moment
        )
        b_uu = self.get_b_uu_finite_differences(
            sample_3d_grid, sample_dipole, sample_magnetic_moment
        )
        # Check if the divergence of B is zero
        npt.assert_allclose(-b_uu, b_ee + b_nn, atol=1e-12)
